import xtrack as xt

d_zeta = 1e-3

collider = xt.Environment.from_json(
    '../../test_data/hllhc15_collider/collider_00_from_mad.json'
)
collider.build_trackers()


line = collider.lhcb1
line.cycle('ip2', inplace=True)
tw4d = line.twiss(method='4d')

tw_plus = line.twiss(method='4d', start=0, end=len(line) - 1,
                     init=xt.TwissInit(element_name=line.element_names[0],
                                             line=line, zeta=d_zeta))
tw_minus = line.twiss(method='4d', start=0, end=len(line) - 1,
                        init=xt.TwissInit(element_name=line.element_names[0],
                                                line=line, zeta=-d_zeta))
dx_zeta_rf_off_crab_off = (tw_plus.x - tw_minus.x)/(tw_plus.zeta - tw_minus.zeta)

tw_plus_closed = line.twiss(method='4d', zeta0=d_zeta)
tw_minus_closed = line.twiss(method='4d', zeta0=-d_zeta)
cl_dx_zeta_rf_off_crab_off = (tw_plus_closed.x - tw_minus_closed.x)/(tw_plus_closed.zeta - tw_minus_closed.zeta)

line.vars['vrf400'] = 16

tw_plus = line.twiss(method='4d', start=0, end=len(line) - 1,
                     init=xt.TwissInit(element_name=line.element_names[0],
                                             line=line, zeta=d_zeta))
tw_minus = line.twiss(method='4d', start=0, end=len(line) - 1,
                        init=xt.TwissInit(element_name=line.element_names[0],
                                                line=line, zeta=-d_zeta))
dx_zeta_4d_rf_on_crab_off = (tw_plus.x - tw_minus.x)/(tw_plus.zeta - tw_minus.zeta)

tw_plus_closed = line.twiss(method='4d', zeta0=d_zeta)
tw_minus_closed = line.twiss(method='4d', zeta0=-d_zeta)
cl_dx_zeta_4d_rf_on_crab_off = (tw_plus_closed.x - tw_minus_closed.x)/(tw_plus_closed.zeta - tw_minus_closed.zeta)

line.vars['on_crab1'] = -190

tw_plus = line.twiss(method='4d', start=0, end=len(line) - 1,
                        init=xt.TwissInit(element_name=line.element_names[0],
                                                line=line, zeta=d_zeta))
tw_minus = line.twiss(method='4d', start=0, end=len(line) - 1,
                        init=xt.TwissInit(element_name=line.element_names[0],
                                                line=line, zeta=-d_zeta))
dx_zeta_4d_rf_on_crab_on = (tw_plus.x - tw_minus.x)/(tw_plus.zeta - tw_minus.zeta)

tw_plus_closed = line.twiss(method='4d', zeta0=d_zeta)
tw_minus_closed = line.twiss(method='4d', zeta0=-d_zeta)
cl_dx_zeta_4d_rf_on_crab_on = (tw_plus_closed.x - tw_minus_closed.x)/(tw_plus_closed.zeta - tw_minus_closed.zeta)

line.vars['vrf400'] = 0
tw_plus = line.twiss(method='4d', start=0, end=len(line) - 1,
                        init=xt.TwissInit(element_name=line.element_names[0],
                                                line=line, zeta=d_zeta))
tw_minus = line.twiss(method='4d', start=0, end=len(line) - 1,
                        init=xt.TwissInit(element_name=line.element_names[0],
                                                line=line, zeta=-d_zeta))
dx_zeta_rf_off_crab_on = (tw_plus.x - tw_minus.x)/(tw_plus.zeta - tw_minus.zeta)

tw_plus_closed = line.twiss(method='4d', zeta0=d_zeta)
tw_minus_closed = line.twiss(method='4d', zeta0=-d_zeta)
cl_dx_zeta_rf_off_crab_on = (tw_plus_closed.x - tw_minus_closed.x)/(tw_plus_closed.zeta - tw_minus_closed.zeta)


import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
plt.plot(tw_plus.s, dx_zeta_4d_rf_on_crab_on, label='rf on, crab on', color='g')
plt.plot(tw_plus.s, dx_zeta_rf_off_crab_on, label='rf off, crab on', color='m')
plt.plot(tw_plus.s, dx_zeta_4d_rf_on_crab_off, label='rf on, crab off', color='b')
plt.plot(tw_plus.s, dx_zeta_rf_off_crab_off, label='rf off, crab off', color='r')
plt.axvline(x=line.get_s_position('ip4'), color='k', linestyle='--', label='ip4')
plt.legend()
plt.xlabel('s [m]')
plt.ylabel('dx/dzeta')





line.vars['vrf400'] = 16

line.vars['on_crab1'] = 0
tw_crab_off = line.twiss()
line.vars['on_crab1'] = -190
tw_crab_on = line.twiss()

plt.figure(2)
plt.plot(tw_crab_on.s, tw_crab_on.dx_zeta, label='crab on')
plt.plot(tw_crab_off.s, tw_crab_off.dx_zeta, label='crab off')
plt.legend()
plt.xlabel('s [m]')
plt.ylabel('dx/dzeta')


plt.figure(10)
plt.plot(tw_plus.s, cl_dx_zeta_4d_rf_on_crab_on, label='rf on, crab on', color='g')
plt.plot(tw_plus.s, cl_dx_zeta_rf_off_crab_on, label='rf off, crab on', color='m')
plt.plot(tw_plus.s, cl_dx_zeta_4d_rf_on_crab_off, label='rf on, crab off', color='b')
plt.plot(tw_plus.s, cl_dx_zeta_rf_off_crab_off, label='rf off, crab off', color='r')
plt.plot(tw_plus.s, tw_crab_on.dx_zeta, '--', label='6d', color='k', linewidth=2)
plt.legend()
plt.xlabel('s [m]')
plt.ylabel('dx/dzeta')

plt.show()





import numpy as np
import xtrack as xt

import xtrack._temp.lhc_match as lm

collider = xt.Environment.from_json('hllhc.json')
collider.build_trackers()
collider.vars.load_madx_optics_file(
    "../../test_data/hllhc15_thick/opt_round_150_1500.madx")

arc_periodic_solution =lm.get_arc_periodic_solution(collider)

for aa in ['12', '23', '34', '45', '56', '67', '78', '81']:
    assert np.isclose(arc_periodic_solution['lhcb1'][aa].mux[-1],
                        collider.vars[f'mux{aa}b1']._value, atol=1e-10, rtol=0)
    assert np.isclose(arc_periodic_solution['lhcb1'][aa].muy[-1],
                        collider.vars[f'muy{aa}b1']._value, atol=1e-10, rtol=0)
    assert np.isclose(arc_periodic_solution['lhcb2'][aa].mux[-1],
                        collider.vars[f'mux{aa}b2']._value, atol=1e-10, rtol=0)
    assert np.isclose(arc_periodic_solution['lhcb2'][aa].muy[-1],
                        collider.vars[f'muy{aa}b2']._value, atol=1e-10, rtol=0)


tw_presq_ip1_b1 = lm.propagate_optics_from_beta_star(collider, ip_name='ip1',
        line_name='lhcb1', start='s.ds.r8.b1', end='e.ds.l2.b1',
        beta_star_x=0.5, beta_star_y = 0.5)

assert np.isclose(
    tw_presq_ip1_b1['mux', 'ip1'] -  tw_presq_ip1_b1['mux', 's.ds.l1.b1'],
    collider.vars['muxip1b1_l']._value, atol=1e-10, rtol=0)
assert np.isclose(
    tw_presq_ip1_b1['muy', 'ip1'] -  tw_presq_ip1_b1['muy', 's.ds.l1.b1'],
    collider.vars['muyip1b1_l']._value, atol=1e-10, rtol=0)
assert np.isclose(
    tw_presq_ip1_b1['mux', 'e.ds.r1.b1'] -  tw_presq_ip1_b1['mux', 'ip1'],
    collider.vars['muxip1b1_r']._value, atol=1e-10, rtol=0)
assert np.isclose(
    tw_presq_ip1_b1['muy', 'e.ds.r1.b1'] -  tw_presq_ip1_b1['muy', 'ip1'],
    collider.vars['muyip1b1_r']._value, atol=1e-10, rtol=0)

# Beam 1

tw_sq_ip1_b1 = lm.propagate_optics_from_beta_star(collider, ip_name='ip1',
        line_name='lhcb1', start='s.ds.r7.b1', end='e.ds.l3.b1',
        beta_star_x=0.15, beta_star_y = 0.15)
vt = collider.vars.get_table()

# Check that the phase advance the whole ATS section is preserved
assert np.isclose(
        tw_sq_ip1_b1['mux', 'e.ds.r2.b1'] - tw_sq_ip1_b1['mux', 's.ds.l8.b1'],
        (vt['value', 'muxip8b1']+vt['value', 'mux81b1']+vt['value', 'muxip1b1']
            + vt['value', 'mux12b1'] +  vt['value', 'muxip2b1']),
        atol=1e-9, rtol=0)

# Change of arc phase advance wrt periodic solution
dmux_81 = ((tw_sq_ip1_b1['mux', 's.ds.l1.b1'] - tw_sq_ip1_b1['mux', 'e.ds.r8.b1'])
           - vt['value', 'mux81b1'])

# Change ip1_l phase advance wrt presquezed solution
dmux_ip1_l = (tw_sq_ip1_b1['mux', 'ip1'] - tw_sq_ip1_b1['mux', 's.ds.l1.b1']
              - vt['value', 'muxip1b1_l'])

muxip8b1_sq = (tw_sq_ip1_b1['mux', 'e.ds.r8.b1'] - tw_sq_ip1_b1['mux', 's.ds.l8.b1'])

assert np.isclose(muxip8b1_sq - vt['value', 'muxip8b1'], -(dmux_81 + dmux_ip1_l),
                  atol=1e-9, rtol=0)


# Beam 2
tw_sq_ip1_b2 = lm.propagate_optics_from_beta_star(collider, ip_name='ip1',
        line_name='lhcb2', start='s.ds.r7.b2', end='e.ds.l3.b2',
        beta_star_x=0.15, beta_star_y = 0.15)










# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import pathlib
import json
import numpy as np

import xobjects as xo
import xtrack as xt

xt.enable_pyheadtail_interface()

fname_line = '../../test_data/lhc_no_bb/line_and_particle.json'

num_turns = int(100)
n_part = 200

####################
# Choose a context #
####################

context = xo.ContextCpu()

##############
# Get a line #
##############

with open(fname_line, 'r') as fid:
     input_data = json.load(fid)
line = xt.Line.from_dict(input_data['line'])


#########################
# Add PyHEADTAIL damper #
#########################

from PyHEADTAIL.feedback.transverse_damper import TransverseDamper
damper = TransverseDamper(dampingrate_x=10., dampingrate_y=15.)
line.append_element(damper, 'Damper')

#################
# Build tracker #
#################
line.build_tracker(_context=context)

######################
# Get some particles #
######################
particles = xt.Particles(_context=context,
                         p0c=6500e9,
                         x=np.random.uniform(-1e-3, 1e-3, n_part)+1e-3,
                         px=np.random.uniform(-1e-7, 1e-7, n_part),
                         y=np.random.uniform(-0.5e-3, 0.5e-3, n_part)-1.2e-3,
                         py=np.random.uniform(-1e-7, 1e-7, n_part),
                         zeta=0*np.random.uniform(-1e-2, 1e-2, n_part),
                         delta=0.*np.random.uniform(-1e-5, 1e-5, n_part)
                         )
#########
# Track #
#########
line.track(particles, num_turns=num_turns, turn_by_turn_monitor=True)


########
# Plot #
########

res = line.record_last_track

import matplotlib.pyplot as plt
plt.close('all')
fig1 = plt.figure(1)
sp1 = fig1.add_subplot(2,1,1)
sp2 = fig1.add_subplot(2,1,2)
sp1.plot(np.mean(res.x, axis=0))
sp2.plot(np.mean(res.y, axis=0))
plt.show()



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import pathlib
import json
import numpy as np

import xobjects as xo
import xtrack as xt


fname_line = '../../test_data/lhc_no_bb/line_and_particle.json'

num_turns = int(100)
n_part = 200

####################
# Choose a context #
####################

context = xo.ContextCpu()
#context = xo.ContextCupy()
#context = xo.ContextPyopencl('0.0')

##############
# Get a line #
##############

with open(fname_line, 'r') as fid:
     input_data = json.load(fid)
line = xt.Line.from_dict(input_data['line'])

##################
# Build TrackJob #
##################
line.build_tracker(_context=context)

######################
# Get some particles #
######################
particles = xt.Particles(_context=context,
                         p0c=6500e9,
                         x=np.random.uniform(-1e-3, 1e-3, n_part),
                         px=np.random.uniform(-1e-5, 1e-5, n_part),
                         y=np.random.uniform(-2e-3, 2e-3, n_part),
                         py=np.random.uniform(-3e-5, 3e-5, n_part),
                         zeta=np.random.uniform(-1e-2, 1e-2, n_part),
                         delta=np.random.uniform(-1e-4, 1e-4, n_part),
                         )
#########
# Track #
#########
line.track(particles, num_turns=num_turns)



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json
import pathlib
import numpy as np

import xobjects as xo
import xtrack as xt

import ducktrack as dtk

test_data_folder = pathlib.Path(
        __file__).parent.joinpath('../../test_data').absolute()

fname_line_particles = test_data_folder.joinpath('lhc_no_bb/line_and_particle.json')
rtol_10turns = 1e-9; atol_10turns=4e-11
test_backtracker=True

#fname_line_particles = test_data_folder.joinpath(
#                                './lhc_with_bb/line_and_particle.json')
#rtol_10turns = 1e-9; atol_10turns=2e-11
#test_backtracker = False

# fname_line_particles = test_data_folder.joinpath(
#                          './hllhc_14/line_and_particle.json')
# rtol_10turns=1e-9; atol_10turns=1e-11 # 2e-10 needed for delta = 1e-3
# test_backtracker = False

#fname_line_particles = test_data_folder.joinpath(
#                    './sps_w_spacecharge/line_with_spacecharge_and_particle.json')
#rtol_10turns = 2e-8; atol_10turns=7e-9
#test_backtracker = False

####################
# Choose a context #
####################

context = xo.ContextCpu()
#context = xo.ContextCupy()
#context = xo.ContextPyopencl('0.0')

#############
# Load file #
#############

with open(fname_line_particles, 'r') as fid:
    input_data = json.load(fid)

##############
# Get a line #
##############

line = xt.Line.from_dict(input_data['line'])

#################
# Build Tracker #
#################
print('Build tracker...')
line.build_tracker(_context=context)
line.reset_s_at_end_turn = False

######################
# Get some particles #
######################
particles = xt.Particles(_context=context, **input_data['particle'])

# To test off momentum one can do the following:
# #particles.delta = 1e-3
# input_data['particle'] = particles.to_dict()

#########
# Track #
#########
print('Track a few turns...')
n_turns = 10
line.track(particles, num_turns=n_turns)

###########################
# Check against ducktrack #
###########################
print('Check against ducktrack...')

testline = dtk.TestLine.from_dict(input_data['line'])

ip_check = 0
vars_to_check = ['x', 'px', 'y', 'py', 'zeta', 'delta', 's']
dtk_part = dtk.TestParticles.from_dict(input_data['particle']).copy()
for _ in range(n_turns):
   testline.track(dtk_part)

for vv in vars_to_check:
    dtk_value = getattr(dtk_part, vv)[0]
    xt_value = context.nparray_from_context_array(getattr(particles, vv))[ip_check]
    passed = np.isclose(xt_value, dtk_value, rtol=rtol_10turns, atol=atol_10turns)

    if not passed:
        print(f'Not passed on var {vv}!\n'
              f'    dtk:   {dtk_value: .7e}\n'
              f'    xtrack: {xt_value: .7e}\n')
        raise ValueError

#####################
# Check backtracker #
#####################

if test_backtracker:
    line.track(particles, num_turns=n_turns, backtrack=True)

    dtk_part = dtk.TestParticles.from_dict(input_data['particle']).copy()

    for vv in vars_to_check:
        dtk_value = getattr(dtk_part, vv)[0]
        xt_value = context.nparray_from_context_array(getattr(particles, vv))[ip_check]
        passed = np.isclose(xt_value, dtk_value, rtol=rtol_10turns,
                            atol=atol_10turns)
        if not passed and vv=='s':
            passed = np.isclose(xt_value, dtk_value,
                    rtol=rtol_10turns, atol=1e-8)

        if not passed:
            print(f'Not passend on backtrack for var {vv}!\n'
                  f'    dtk:   {dtk_value: .7e}\n'
                  f'    xtrack: {xt_value: .7e}\n')
            raise ValueError

##############
# Check  ebe #
##############
print('Check element-by-element against ducktrack...')
dtk_part = dtk.TestParticles.from_dict(input_data['particle']).copy()
vars_to_check = ['x', 'px', 'y', 'py', 'zeta', 'delta', 's']
problem_found = False
diffs = []
s_coord = []
for ii, (eedtk, nn) in enumerate(zip(testline.elements, testline.element_names)):
    vars_before = {vv: getattr(dtk_part, vv)[0] for vv in vars_to_check}
    particles = xt.Particles.from_dict(dtk_part.to_dict(), _context=context)

    line.track(particles, ele_start=ii, num_elements=1)

    eedtk.track(dtk_part)
    s_coord.append(dtk_part.s[0])
    diffs.append([])
    for vv in vars_to_check:
        dtk_change = getattr(dtk_part, vv)[0] - vars_before[vv]
        xt_change = (context.nparray_from_context_array(
                getattr(particles, vv))[ip_check]- vars_before[vv])
        passed = np.isclose(xt_change, dtk_change, rtol=1e-10, atol=5e-14)
        if not passed:
            problem_found = True
            print(f'Not passend on var {vv}!\n'
                  f'    dtk:   {dtk_change: .7e}\n'
                  f'    xtrack: {xt_change: .7e}\n')
            break
        diffs[-1].append(np.abs(
            context.nparray_from_context_array(
                getattr(particles, vv))[ip_check] - getattr(dtk_part, vv)[0]))

    if not passed:
        print(f'\nelement {nn}')
        break

    if test_backtracker:
        line.track(particles, ele_start=ii, ele_stop=ii+1, backtrack=True)

        for vv in vars_to_check:
            xt_value = context.nparray_from_context_array(
                                        getattr(particles, vv))[ip_check]
            passed = np.isclose(xt_value, vars_before[vv],
                                rtol=1e-10, atol=1e-13)
            if not passed:
                problem_found = True
                print(f'\nNot passend on var {vv}!\n'
                      f'    before: {vars_before[vv]: .7e}\n'
                      f'    xtrack: {xt_value: .7e}\n')
                break
        if not passed:
            print(f'\nelement {nn}')
            break

    print(f'Check passed for element: {nn}              ', end='\r', flush=True)


diffs = np.array(diffs)

if not problem_found:
    print('\nAll passed on context:')
    print(context)

import matplotlib.pyplot as plt
plt.close('all')
fig = plt.figure(1, figsize=(6.4*1.5, 4.8*1.3))
for ii_plt, (vv, uu) in enumerate(
        zip(['x', 'px', 'y', 'py', r'$\zeta$', r'$\delta$'],
            ['[m]', '[-]', '[m]', '[-]', '[m]', '[-]'])):
    ax = fig.add_subplot(3, 2, ii_plt+1)
    ax.plot(s_coord, diffs[:, ii_plt])
    ax.set_ylabel('Difference on '+ vv + ' ' + uu)
    ax.set_xlabel('s [m]')
fig.subplots_adjust(hspace=.48)
plt.show()





# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xtrack as xt

fname_line_particles = '../../test_data/hllhc15_noerrors_nobb/line_and_particle.json'
line = xt.Line.from_json(fname_line_particles)
line.particle_ref = xt.Particles(p0c=7e12, mass0=xt.PROTON_MASS_EV)
line.build_tracker()



tw= line.twiss()
W_before_propagation, _, _, _= xt.linear_normal_form.compute_linear_normal_form(tw.R_matrix)
# tw_full_inverse = line.twiss(use_full_inverse=True)

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
for jj in range(6):
    plt.figure(jj+1)
    plt.plot(tw.W_matrix[0][:, jj])
    plt.plot(tw.W_matrix[-1][:, jj])
    plt.plot(W_before_propagation[:, jj], 'x')
    # plt.plot(tw_full_inverse.W_matrix[0][:, jj])
    # plt.plot(tw_full_inverse.W_matrix[-1][:, jj])


plt.show()





#
# This example creates a simple line with 8 quadrupoles separated by drifts.
# Each quadrupole is split into two thin multipoles, separated by a small drift.
# The first half has one optic and the second half another optic.
#
# Knobs are set for the quadrupoles, and then the second, third, fourth and fifth quads
# are used to match the optics between the two halves.
#

import numpy as np
import xtrack as xt
import xobjects as xo
import matplotlib.pyplot as plt

#### create one FODO line with different optics in first and second half ####
kf1 = 0.05971 / 2
kd1 = -0.05971 / 2
ld1 = 33
lq1 = 1
kf2 = 0.084445 / 2
kd2 = -0.084445 / 2
ld2 = 33
lq2 = 1

line = xt.Line(
    elements=[
        xt.Marker(),
        xt.Drift(length=ld1 / 2),
        xt.Multipole(knl=[0, kf1 / 2], ksl=[0, 0]),
        xt.Drift(length=lq1 / 2),
        xt.Marker(),
        xt.Drift(length=lq1 / 2),
        xt.Multipole(knl=[0, kf1 / 2], ksl=[0, 0]),  # MQD
        xt.Drift(length=ld1),
        xt.Multipole(knl=[0, kd1 / 2], ksl=[0, 0]),
        xt.Drift(length=lq1 / 2),
        xt.Marker(),
        xt.Drift(length=lq1 / 2),
        xt.Multipole(knl=[0, kd1 / 2], ksl=[0, 0]),  # MQD
        xt.Drift(length=ld1),
        xt.Multipole(knl=[0, kf1 / 2], ksl=[0, 0]),
        xt.Drift(length=lq1 / 2),
        xt.Marker(),
        xt.Drift(length=lq1 / 2),
        xt.Multipole(knl=[0, kf1 / 2], ksl=[0, 0]),  # MQD
        xt.Drift(length=ld1),
        xt.Multipole(knl=[0, kd1 / 2], ksl=[0, 0]),
        xt.Drift(length=lq1 / 2),
        xt.Marker(),
        xt.Drift(length=lq1 / 2),
        xt.Multipole(knl=[0, kd1 / 2], ksl=[0, 0]),  # MQF
        xt.Drift(length=ld1 / 2 + ld2 / 2),
        xt.Multipole(knl=[0, kf2 / 2], ksl=[0, 0]),
        xt.Drift(length=lq2 / 2),
        xt.Marker(),
        xt.Drift(length=lq2 / 2),
        xt.Multipole(knl=[0, kf2 / 2], ksl=[0, 0]),  # MQD
        xt.Drift(length=ld2),
        xt.Multipole(knl=[0, kd2 / 2], ksl=[0, 0]),
        xt.Drift(length=lq2 / 2),
        xt.Marker(),
        xt.Drift(length=lq2 / 2),
        xt.Multipole(knl=[0, kd2 / 2], ksl=[0, 0]),  # MQD
        xt.Drift(length=ld2),
        xt.Multipole(knl=[0, kf2 / 2], ksl=[0, 0]),
        xt.Drift(length=lq2 / 2),
        xt.Marker(),
        xt.Drift(length=lq2 / 2),
        xt.Multipole(knl=[0, kf2 / 2], ksl=[0, 0]),  # MQD
        xt.Drift(length=ld2),
        xt.Multipole(knl=[0, kd2 / 2], ksl=[0, 0]),
        xt.Drift(length=lq2 / 2),
        xt.Marker(),
        xt.Drift(length=lq2 / 2),
        xt.Multipole(knl=[0, kd2 / 2], ksl=[0, 0]),  # MQF
        xt.Drift(length=ld2 / 2),
        xt.Marker(),
    ],
    element_names=[
        "start",
        "drift0",
        "mq0..1",
        "drift01",
        "mq0",
        "drift02",
        "mq0..2",
        "drift1",
        "mq1..1",
        "drift11",
        "mq1",
        "drift12",
        "mq1..2",
        "drift2",
        "mq2..1",
        "drift21",
        "mq2",
        "drift22",
        "mq2..2",
        "drift3",
        "mq3..1",
        "drift31",
        "mq3",
        "drift32",
        "mq3..2",
        "drift4",
        "mq4..1",
        "drift41",
        "mq4",
        "drift42",
        "mq4..2",
        "drift5",
        "mq5..1",
        "drift51",
        "mq5",
        "drift52",
        "mq5..2",
        "drift6",
        "mq6..1",
        "drift61",
        "mq6",
        "drift62",
        "mq6..2",
        "drift7",
        "mq7..1",
        "drift71",
        "mq7",
        "drift72",
        "mq7..2",
        "drift8",
        "end",
    ],
)
line.particle_ref = xt.Particles(p0c=450e9, q0=1, mass0=xt.PROTON_MASS_EV)

context = xo.ContextCpu()
line.build_tracker(_context=context)

#### calculate matched twiss parameters ####
Rqf1 = np.array([[1, 0], [kf1 / 2, 1]])
Rqd1 = np.array([[1, 0], [kd1 / 2, 1]])
Rd1 = np.array([[1, ld1], [0, 1]])
Rdmq1 = np.array([[1, lq1 / 2], [0, 1]])
Rcell1f = (
    Rdmq1.dot(Rqf1)
    .dot(Rd1)
    .dot(Rqd1)
    .dot(Rdmq1)
    .dot(Rdmq1)
    .dot(Rqd1)
    .dot(Rd1)
    .dot(Rqf1)
    .dot(Rdmq1)
)
Rcell1d = (
    Rdmq1.dot(Rqd1)
    .dot(Rd1)
    .dot(Rqf1)
    .dot(Rdmq1)
    .dot(Rdmq1)
    .dot(Rqf1)
    .dot(Rd1)
    .dot(Rqd1)
    .dot(Rdmq1)
)
Rqf2 = np.array([[1, 0], [kf2 / 2, 1]])
Rqd2 = np.array([[1, 0], [kd2 / 2, 1]])
Rd2 = np.array([[1, ld2], [0, 1]])
Rdmq2 = np.array([[1, lq2 / 2], [0, 1]])
Rcell2f = (
    Rdmq2.dot(Rqf2)
    .dot(Rd2)
    .dot(Rqd2)
    .dot(Rdmq2)
    .dot(Rdmq2)
    .dot(Rqd2)
    .dot(Rd2)
    .dot(Rqf2)
    .dot(Rdmq2)
)
Rcell2d = (
    Rdmq2.dot(Rqd2)
    .dot(Rd2)
    .dot(Rqf2)
    .dot(Rdmq2)
    .dot(Rdmq2)
    .dot(Rqf2)
    .dot(Rd2)
    .dot(Rqd2)
    .dot(Rdmq2)
)
mu1 = np.arccos(0.5 * (Rcell1f[0, 0] + Rcell1f[1, 1]))
betMax1 = np.sqrt(
    np.abs(Rcell1d[0, 1] * Rcell1d[0, 0] / (Rcell1d[1, 0] * Rcell1d[1, 1]))
)
betMin1 = np.sqrt(
    np.abs(Rcell1f[0, 1] * Rcell1f[1, 1] / (Rcell1f[0, 0] * Rcell1f[1, 0]))
)
mu2 = np.arccos(0.5 * (Rcell2f[0, 0] + Rcell2f[1, 1]))
betMax2 = np.sqrt(
    np.abs(Rcell2d[0, 1] * Rcell2d[0, 0] / (Rcell2d[1, 0] * Rcell2d[1, 1]))
)
betMin2 = np.sqrt(
    np.abs(Rcell2f[0, 1] * Rcell2f[1, 1] / (Rcell2f[0, 0] * Rcell2f[1, 0]))
)

print(
    f"first half: mu: {mu1*180/np.pi:.2f} -- betMax: {betMax1:.2f} -- betMin: {betMin1:.2f}"
)
print(
    f"second half: mu: {mu2*180/np.pi:.2f} -- betMax: {betMax2:.2f} -- betMin: {betMin2:.2f}"
)

#### check twiss ####
tw1s = line.twiss(
    start="mq0",
    end="mq3",
    init=xt.TwissInit(betx=betMax1, bety=betMin1, element_name="mq0", line=line),
)
tw2s = line.twiss(
    start="mq4",
    end="mq7",
    init=xt.TwissInit(betx=betMax2, bety=betMin2, element_name="mq4", line=line),
)
print(tw1s.cols["s", "betx", "bety", "alfx", "alfy"])
print(tw2s.cols["s", "betx", "bety", "alfx", "alfy"])

#### create knobs for the quadrupole strengths ####
line._init_var_management()
kfd_refs = [kf1, kd1, kf1, kd1, kf2, kd2, kf2, kd2]
for i, kfd in enumerate(kfd_refs):
    line.vv["kq" + str(i)] = kfd / 2
    line.element_refs["mq" + str(i) + "..1"].knl[1] = line.vars["kq" + str(i)]
    line.element_refs["mq" + str(i) + "..2"].knl[1] = line.vars["kq" + str(i)]


#### use mq2, mq3, mq4 and mq5 to match the optics between the two halves
opt = line.match(
    solve=False,
    default_tol={None: 5e-10},
    init=tw1s.get_twiss_init("mq0"),
    start=("mq0"),
    end=("end"),
    targets=[
        xt.Target(tar="alfx", value=tw2s.rows["mq6"]["alfx"][0], at="mq6"),
        xt.Target(tar="alfy", value=tw2s.rows["mq6"]["alfy"][0], at="mq6"),
        xt.Target(tar="betx", value=tw2s.rows["mq6"]["betx"][0], at="mq6"),
        xt.Target(tar="bety", value=tw2s.rows["mq6"]["bety"][0], at="mq6"),
    ],
    vary=xt.VaryList(["kq2", "kq3", "kq4", "kq5"]),
)
opt.solve()
print(opt.log())

#### cross-check the final twiss ####
tw3s = line.twiss(
    start="mq0",
    end="end",
    init=xt.TwissInit(betx=betMax1, bety=betMin1, element_name="mq0", line=line),
)
print(tw3s.cols["s", "betx", "bety", "alfx", "alfy"])



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json

import xtrack as xt
import xpart as xp
import xobjects as xo

context = xo.ContextCpu()

with open('../../test_data/hllhc15_noerrors_nobb/line_and_particle.json') as f:
    dct = json.load(f)
line = xt.Line.from_dict(dct['line'])
line.particle_ref = xt.Particles.from_dict(dct['particle'])

line.build_tracker()

num_particles = 50
particles = xp.generate_matched_gaussian_bunch(line=line,
                                               num_particles=num_particles,
                                               nemitt_x=2.5e-6,
                                               nemitt_y=2.5e-6,
                                               sigma_z=9e-2)

num_turns = 100
monitor = xt.ParticlesMonitor(_context=context,
                              start_at_turn=5, stop_at_turn=10,
                              n_repetitions=3,      # <--
                              repetition_period=20, # <--
                              num_particles=num_particles)
for iturn in range(num_turns):
    monitor.track(particles)
    line.track(particles)

# monitor contains the measured data. For all particles
# variables the first index provides the frame index.
# For example, monitor.x[0, :, :] contains the recorded
# x position for the turns 5 to 10, monitor.x[1, :, :]
# contains the recorded x position for the turns 25 to 30, etc.
# The turn indeces that are recorded can be inspected in
# monitor.at_turn.



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
from scipy.constants import e as qe
from scipy.constants import c as clight

from ..base_classes import Element
from .gaussian_fields import get_Ex_Ey_Gx_Gy_gauss
from . import BB6D
from . import BB6Ddata


class BeamBeam4D(Element):
    """Interaction with a transverse-Gaussian strong beam (4D modelling)."""

    _description = [
        (
            "charge",
            "e",
            "Charge of the interacting distribution (strong beam)",
            0,
        ),
        ("sigma_x", "m", "Horizontal size of the strong beam (r.m.s.)", 1.0),
        ("sigma_y", "m", "Vertical size of the strong beam (r.m.s.)", 1.0),
        ("beta_r", "", "Relativistic beta of the stron beam", 1.0),
        (
            "x_bb",
            "m",
            "H. position of the strong beam w.r.t. the reference trajectory",
            0,
        ),
        (
            "y_bb",
            "m",
            "V. position of the strong beam w.r.t. the reference trajectory",
            0,
        ),
        ("d_px", "", "H. kick subtracted after the interaction.", 0),
        ("d_py", "", "V. kick subtracted after the interaction.", 0),
    ]

    _extra = [
        ("min_sigma_diff", "m", "Threshold to detect round beam", 1e-28),
        ("enabled", "", "Switch for closed orbit search", True),
    ]

    @classmethod
    def from_dict(cls, dct):
        if 'other_beam_Sigma_11' in dct.keys():
            # Comes from xfields
            assert dct.get('other_beam_Sigma_13', 0) == 0
            assert dct.get('ref_shift_x', 0) == 0
            assert dct.get('ref_shift_y', 0) == 0

            kwargs = {
            "charge": dct["other_beam_q0"]*dct['other_beam_num_particles'],
            "sigma_x": np.sqrt(dct["other_beam_Sigma_11"]),
            "sigma_y": np.sqrt(dct["other_beam_Sigma_33"]),
            "beta_r": dct["other_beam_beta0"],
            "x_bb": dct["other_beam_shift_x"],
            "y_bb": dct["other_beam_shift_y"],
            "d_px": dct["post_subtract_px"],
            "d_py": dct['post_subtract_py'],
            }
        else:
            kwargs = dct

        return cls(**kwargs)

    def track(self, p):
        if self.enabled:
            charge = p.charge_ratio * p.q0
            x = p.x - self.x_bb
            px = p.px
            y = p.y - self.y_bb
            py = p.py

            chi = p.chi

            beta = p.beta0 / p.rvv
            p0c = p.p0c * qe

            Ex, Ey = get_Ex_Ey_Gx_Gy_gauss(
                x,
                y,
                self.sigma_x,
                self.sigma_y,
                min_sigma_diff=1e-10,
                skip_Gs=True,
                mathlib=p._m,
            )

            fact_kick = (
                chi
                * self.charge
                * qe
                * charge
                * qe
                * (1.0 + beta * self.beta_r)
                / (p0c * (beta + self.beta_r))
            )

            px += fact_kick * Ex - self.d_px
            py += fact_kick * Ey - self.d_py

            p.px = px
            p.py = py


class BeamBeam6D(Element):
    """Interaction with a transverse-Gaussian strong beam (6D modelling).

    http://cds.cern.ch/record/2306400
    """

    _description = [
        (
            "phi",
            "rad",
            "Crossing angle (>0 weak beam increases"
            "x in the direction motion)",
            0,
        ),
        ("alpha", "rad", "Crossing plane tilt angle (>0 x tends to y)", 0),
        (
            "x_bb_co",
            "m",
            "H. position of the strong beam w.r.t. the closed orbit",
            0,
        ),
        (
            "y_bb_co",
            "m",
            "V. position of the strong beam w.r.t. the closed orbit",
            0,
        ),
        (
            "charge_slices",
            "qe",
            "Charge of the interacting slices (strong beam)",
            (0.0),
        ),
        (
            "zeta_slices",
            "m",
            "Longitudinal position of the interacting"
            "slices (>0 head of the strong).",
            (0.0),
        ),
        (
            "sigma_11",
            "m^2",
            "Sigma_11 element of the sigma matrix of the strong beam",
            1.0,
        ),
        (
            "sigma_12",
            "m",
            "Sigma_12 element of the sigma matrix of the strong beam",
            0,
        ),
        (
            "sigma_13",
            "m^2",
            "Sigma_13 element of the sigma matrix of the strong beam",
            0,
        ),
        (
            "sigma_14",
            "m",
            "Sigma_14 element of the sigma matrix of the strong beam",
            0,
        ),
        (
            "sigma_22",
            "",
            "Sigma_22 element of the sigma matrix of the strong beam",
            0,
        ),
        (
            "sigma_23",
            "m",
            "Sigma_23 element of the sigma matrix of the strong beam",
            0,
        ),
        (
            "sigma_24",
            "",
            "Sigma_24 element of the sigma matrix of the strong beam",
            0,
        ),
        (
            "sigma_33",
            "m^2",
            "Sigma_33 element of the sigma matrix of the strong beam",
            1.0,
        ),
        (
            "sigma_34",
            "m",
            "Sigma_34 element of the sigma matrix of the strong beam",
            0,
        ),
        (
            "sigma_44",
            "",
            "Sigma_44 element of the sigma matrix of the strong beam",
            0,
        ),
        ("x_co", "m", "x coordinate the closed orbit (weak beam).", 0),
        ("px_co", "", "px coordinate the closed orbit (weak beam).", 0),
        ("y_co", "m", "y coordinate the closed orbit (weaek beam).", 0),
        ("py_co", "", "py coordinate the closed orbit (weaek beam).", 0),
        ("zeta_co", "m", "zeta coordinate the closed orbit (weaek beam).", 0),
        ("delta_co", "", "delta coordinate the closed orbit (weaek beam).", 0),
        ("d_x", "m", "Quantity subtracted from x after the interaction.", 0),
        ("d_px", "", "Quantity subtracted from px after the interaction.", 0),
        ("d_y", "m", "Quantity subtracted from y after the interaction.", 0),
        ("d_py", "", "Quantity subtracted from py after the interaction.", 0),
        (
            "d_zeta",
            "m",
            "Quantity subtracted from sigma after the interaction.",
            0,
        ),
        (
            "d_delta",
            "",
            "Quantity subtracted from delta after the interaction.",
            0,
        ),
    ]
    _extra = [
        ("min_sigma_diff", "m", "Threshold to detect round beam", 1e-28),
        (
            "threshold_singular",
            "",
            "Threshold to detect small denominator",
            1e-28,
        ),
        ("enabled", "", "Switch for closed orbit search", True),
    ]

    @classmethod
    def from_dict(cls, dct):

        if 'slices_other_beam_Sigma_11_star' in dct.keys():
            # Comes from xfields
            x_st=np.array(dct['slices_other_beam_x_center_star'])
            y_st=np.array(dct['slices_other_beam_y_center_star'])
            zeta_st=np.array(dct['slices_other_beam_zeta_center_star'])

            (_, _, _, _, z_slices, _) = _python_inv_boost_scalar(
                    x_st=x_st,
                    px_st=0*x_st,
                    y_st=y_st,
                    py_st=0*x_st,
                    sigma_st=zeta_st,
                    delta_st=0*x_st,
                    sphi=dct['_sin_phi'],
                    cphi=dct['_cos_phi'],
                    tphi=dct['_tan_phi'],
                    salpha=dct['_sin_alpha'],
                    calpha=dct['_cos_alpha'])
            kwargs = {
                'phi': np.angle(dct['_cos_phi'] + 1j * dct['_sin_phi']),
                'alpha': np.angle(dct['_cos_alpha'] + 1j * dct['_sin_alpha']),
                'x_bb_co': dct['other_beam_shift_x'],
                'y_bb_co': dct['other_beam_shift_y'],
                'charge_slices': np.array(dct['slices_other_beam_num_particles']),
                'zeta_slices': z_slices,
                'sigma_11': dct['slices_other_beam_Sigma_11_star'][0],
                'sigma_12': dct['slices_other_beam_Sigma_12_star'][0]*dct['_cos_phi'],
                'sigma_13': dct['slices_other_beam_Sigma_13_star'][0],
                'sigma_14': dct['slices_other_beam_Sigma_14_star'][0]*dct['_cos_phi'],
                'sigma_22': dct['slices_other_beam_Sigma_22_star'][0]*dct['_cos_phi']*dct['_cos_phi'],
                'sigma_23': dct['slices_other_beam_Sigma_23_star'][0]*dct['_cos_phi'],
                'sigma_24': dct['slices_other_beam_Sigma_24_star'][0]*dct['_cos_phi']*dct['_cos_phi'],
                'sigma_33': dct['slices_other_beam_Sigma_33_star'][0],
                'sigma_34': dct['slices_other_beam_Sigma_34_star'][0]*dct['_cos_phi'],
                'sigma_44': dct['slices_other_beam_Sigma_44_star'][0]*dct['_cos_phi']*dct['_cos_phi'],
                'x_co': dct['ref_shift_x'],
                'px_co': dct['ref_shift_px'],
                'y_co': dct['ref_shift_y'],
                'py_co': dct['ref_shift_py'],
                'zeta_co': dct['ref_shift_zeta'],
                'delta_co': dct['ref_shift_pzeta'],
                'd_x': dct['post_subtract_x'],
                'd_px': dct['post_subtract_px'],
                'd_y': dct['post_subtract_y'],
                'd_py': dct['post_subtract_py'],
                'd_zeta': dct['post_subtract_zeta'],
                'd_delta': dct['post_subtract_pzeta'],
                'min_sigma_diff': dct['min_sigma_diff'],
                'threshold_singular': dct['threshold_singular'],
                'enabled': True
                }
        else:
            kwargs = dct

        return cls(**kwargs)

    def track(self, p):
        if self.enabled:
            bb6data = BB6Ddata.BB6D_init(
                qe,
                self.phi,
                self.alpha,
                self.x_bb_co,
                self.y_bb_co,
                np.atleast_1d(self.charge_slices),
                np.atleast_1d(self.zeta_slices),
                self.sigma_11,
                self.sigma_12,
                self.sigma_13,
                self.sigma_14,
                self.sigma_22,
                self.sigma_23,
                self.sigma_24,
                self.sigma_33,
                self.sigma_34,
                self.sigma_44,
                self.x_co,
                self.px_co,
                self.y_co,
                self.py_co,
                self.zeta_co,
                self.delta_co,
                self.min_sigma_diff,
                self.threshold_singular,
                self.d_x,
                self.d_px,
                self.d_y,
                self.d_py,
                self.d_zeta,
                self.d_delta,
                self.enabled,
            )
            self._bb6ddata = bb6data
            (
                x_ret,
                px_ret,
                y_ret,
                py_ret,
                zeta_ret,
                delta_ret,
            ) = BB6D.BB6D_track(
                p.x,
                p.px,
                p.y,
                p.py,
                p.zeta,
                p.pzeta,
                p.q0 * qe,
                p.p0c / clight * qe,
                bb6data,
                mathlib=p._m,
            )
            self._last_bb6data = bb6data
            p.x = x_ret
            p.px = px_ret
            p.y = y_ret
            p.py = py_ret
            p.zeta = zeta_ret
            if hasattr(p, 'update_delta'):
                p._update_delta(delta_ret)
            else:
                p.delta = delta_ret


def _python_inv_boost_scalar(x_st, px_st, y_st, py_st, sigma_st, delta_st,
                  sphi, cphi, tphi, salpha, calpha):

    pz_st = np.sqrt(
        (1.0 + delta_st) * (1.0 + delta_st) - px_st * px_st - py_st * py_st
    )
    hx_st = px_st / pz_st
    hy_st = py_st / pz_st
    hsigma_st = 1.0 - (delta_st + 1) / pz_st

    Det_L = (
        1.0 / cphi
        + (hx_st * calpha + hy_st * salpha - hsigma_st * sphi) * tphi
    )

    Linv_11 = (
        1.0 / cphi + salpha * tphi * (hy_st - hsigma_st * salpha * sphi)
    ) / Det_L
    Linv_12 = (salpha * tphi * (hsigma_st * calpha * sphi - hx_st)) / Det_L
    Linv_13 = (
        -tphi
        * (
            calpha
            - hx_st * salpha * salpha * sphi
            + hy_st * calpha * salpha * sphi
        )
        / Det_L
    )

    Linv_21 = (calpha * tphi * (-hy_st + hsigma_st * salpha * sphi)) / Det_L
    Linv_22 = (
        1.0 / cphi + calpha * tphi * (hx_st - hsigma_st * calpha * sphi)
    ) / Det_L
    Linv_23 = (
        -tphi
        * (
            salpha
            - hy_st * calpha * calpha * sphi
            + hx_st * calpha * salpha * sphi
        )
        / Det_L
    )

    Linv_31 = -hsigma_st * calpha * sphi / Det_L
    Linv_32 = -hsigma_st * salpha * sphi / Det_L
    Linv_33 = (1.0 + hx_st * calpha * sphi + hy_st * salpha * sphi) / Det_L

    x_i = Linv_11 * x_st + Linv_12 * y_st + Linv_13 * sigma_st
    y_i = Linv_21 * x_st + Linv_22 * y_st + Linv_23 * sigma_st
    sigma_i = Linv_31 * x_st + Linv_32 * y_st + Linv_33 * sigma_st

    h = (delta_st + 1.0 - pz_st) * cphi * cphi

    px_i = px_st * cphi + h * calpha * tphi
    py_i = py_st * cphi + h * salpha * tphi

    delta_i = (
        delta_st
        + px_i * calpha * tphi
        + py_i * salpha * tphi
        - h * tphi * tphi
    )

    return x_i, px_i, y_i, py_i, sigma_i, delta_i

_python_inv_boost = np.vectorize(_python_inv_boost_scalar,
    excluded=("sphi", "cphi", "tphi", "salpha", "calpha"))




import json
import numpy as np
import xtrack as xt

case_name = 'clic_dr'
filename = '../../test_data/clic_dr/line_for_taper.json'
configs = [
    {'radiation_method': 'full', 'p0_correction': False, 'cavity_preserve_angle': False, 'beta_rtol': 2e-2, 'q_atol': 5e-4},
    {'radiation_method': 'full', 'p0_correction': True, 'cavity_preserve_angle': False, 'beta_rtol': 2e-2, 'q_atol': 5e-4},
    {'radiation_method': 'full', 'p0_correction': True, 'cavity_preserve_angle': True, 'beta_rtol': 2e-5, 'q_atol': 5e-4},
    {'radiation_method': 'kick_as_co', 'p0_correction': True, 'cavity_preserve_angle': True, 'beta_rtol': 1e-3, 'q_atol': 5e-4},
    {'radiation_method': 'scale_as_co', 'p0_correction': True, 'cavity_preserve_angle': True, 'beta_rtol': 1e-5, 'q_atol': 5e-4},
]

# case_name = 'fcc-ee'
# filename = 'line_no_radiation.json'
# configs = [
#     {'radiation_method': 'full', 'p0_correction': False, 'cavity_preserve_angle': False, 'beta_rtol': 1e-2, 'q_atol': 5e-4},
#     {'radiation_method': 'full', 'p0_correction': False, 'cavity_preserve_angle': False, 'beta_rtol': 1e-2, 'q_atol': 5e-4, 'delta_in_beta':True},
#     {'radiation_method': 'full', 'p0_correction': True, 'cavity_preserve_angle': False, 'beta_rtol': 5e-3, 'q_atol': 5e-4},
#     {'radiation_method': 'full', 'p0_correction': True, 'cavity_preserve_angle': True, 'beta_rtol': 3e-4, 'q_atol': 5e-4},
#     {'radiation_method': 'kick_as_co', 'p0_correction': True, 'cavity_preserve_angle': True, 'beta_rtol': 3e-3, 'q_atol': 7e-4},
#     {'radiation_method': 'scale_as_co', 'p0_correction': True, 'cavity_preserve_angle': True, 'beta_rtol': 1e-4, 'q_atol': 1e-4},
# ]


with open(filename, 'r') as f:
    line = xt.Line.from_dict(json.load(f))

line.build_tracker()


# Initial twiss (no radiation)
line.configure_radiation(model=None)
tw_no_rad = line.twiss(method='4d', freeze_longitudinal=True)

# Enable radiation
line.configure_radiation(model='mean')
# - Set cavity lags to compensate energy loss
# - Taper magnet strengths
line.compensate_radiation_energy_loss(record_iterations=True)

import matplotlib.pyplot as plt
plt.close('all')
ifig = 0
for conf in configs:

    ifig += 1

    extra_kwargs = {}
    if conf['cavity_preserve_angle']:
        extra_kwargs['matrix_stability_tol'] = 0.1

    # Twiss(es) with radiation
    line.config.XTRACK_CAVITY_PRESERVE_ANGLE = conf['cavity_preserve_angle']
    tw = line.twiss(radiation_method=conf['radiation_method'],
            eneloss_and_damping=(conf['radiation_method'] != 'kick_as_co'),
            **extra_kwargs)
    line.config.XTRACK_CAVITY_PRESERVE_ANGLE = False

    if conf['p0_correction']:
        p0corr = 1 + tw.delta
    else:
        p0corr = 1

    plt.figure(ifig, figsize=(6.4*1.3, 4.8))
    plt.suptitle(f"Radiation method: {conf['radiation_method']}, "
                    f"p0 correction: {conf['p0_correction']}, "
                    f"cavity preserve angle: {conf['cavity_preserve_angle']}")

    betx_beat = tw.betx*p0corr/tw_no_rad.betx-1
    bety_beat = tw.bety*p0corr/tw_no_rad.bety-1
    max_betx_beat = np.max(np.abs(betx_beat))
    max_bety_beat = np.max(np.abs(bety_beat))
    spx = plt.subplot(2,1,1)
    plt.title(f'error on Qx: {abs(tw.qx - tw_no_rad.qx):.2e}     '
                r'$(\Delta \beta_x / \beta_x)_{max}$ = '
                f'{max_betx_beat:.2e}')
    plt.plot(tw.s, betx_beat)
    if 'delta_in_beta' in conf:
        plt.plot(tw.s, -line.delta_taper, 'k')
    plt.ylabel(r'$\Delta \beta_x / \beta_x$')
    plt.ylim(np.max([0.01, 1.1 * max_betx_beat])*np.array([-1, 1]))
    plt.xlim([0, tw.s[-1]])

    plt.subplot(2,1,2, sharex=spx)
    plt.title(f'error on Qy: {abs(tw.qy - tw_no_rad.qy):.2e}     '
                r'$(\Delta \beta_y / \beta_y)_{max}$ = '
                f'{max_bety_beat:.2e}')
    plt.plot(tw.s, bety_beat)
    if 'delta_in_beta' in conf:
        plt.plot(tw.s, -line.delta_taper, 'k')
    plt.ylabel(r'$\Delta \beta_y / \beta_y$')
    plt.ylim(np.max([0.01, 1.1 * max_bety_beat])*np.array([-1, 1]))
    plt.xlabel('s [m]')

    plt.subplots_adjust(hspace=0.35, top=.85)

    plt.savefig(f'./{case_name}_fig{ifig}.png', dpi=200)

    assert np.isclose(line.attr['delta_taper'][0], 0, rtol=0, atol=1e-10)
    assert np.isclose(line.attr['delta_taper'][-1], 0, rtol=0, atol=1e-10)

    assert np.isclose(tw.qx, tw_no_rad.qx, rtol=0, atol=conf['q_atol'])
    assert np.isclose(tw.qy, tw_no_rad.qy, rtol=0, atol=conf['q_atol'])

    assert np.isclose(tw.dqx, tw_no_rad.dqx, rtol=0, atol=1.5e-2*tw.qx)
    assert np.isclose(tw.dqy, tw_no_rad.dqy, rtol=0, atol=1.5e-2*tw.qy)

    assert np.allclose(tw.x, tw_no_rad.x, rtol=0, atol=1e-7)
    assert np.allclose(tw.y, tw_no_rad.y, rtol=0, atol=1e-7)

    assert np.allclose(tw.betx*p0corr, tw_no_rad.betx, rtol=conf['beta_rtol'], atol=0)
    assert np.allclose(tw.bety*p0corr, tw_no_rad.bety, rtol=conf['beta_rtol'], atol=0)

    assert np.allclose(tw.dx, tw.dx, rtol=0.0, atol=0.1e-3)

    assert np.allclose(tw.dy, tw.dy, rtol=0.0, atol=0.1e-3)

plt.figure(100)
for i_iter, mon in enumerate(line._tapering_iterations):
    plt.plot(mon.s.T, mon.delta.T,
             label=f'iter {i_iter} - Ene. loss: {-mon.delta[-1, -1]*100:.2f} %')
    plt.legend(loc='lower left')
    plt.grid(True)
    plt.ylim(top=0.01)
    plt.xlim(left=0, right=mon.s[-1, -1])
    plt.xlabel('s [m]')
    plt.ylabel(r'$\delta$')
    plt.savefig(f'./{case_name}_taper_iter_{i_iter}.png', dpi=200)

plt.show()



import numpy as np

import xtrack as xt
import xobjects as xo
from xtrack.beam_elements.magnets import Magnet

p0 = xt.Particles(kinetic_energy0=50e6,
                  x=[1e-3, -1e-3], y=2e-3, zeta=1e-2, px=10e-3, py=20e-3, delta=1e-2)


# Test entry linear edge alone
bb = xt.Bend(h=0.1, k0=0.11, length=0,
             edge_entry_angle=0.02, edge_exit_angle=0.03,
             edge_entry_hgap=0.04, edge_exit_hgap=0.05,
             edge_entry_fint=0.1, edge_exit_fint=0.2)

bb.edge_entry_active = 1
bb.edge_exit_active = 0
bb.model = 'rot-kick-rot'
bb.num_multipole_kicks = 10


mm = Magnet(h=0.1, k0=0.11, length=0,
             edge_entry_angle=0.02, edge_exit_angle=0.03,
             edge_entry_hgap=0.04, edge_exit_hgap=0.05,
             edge_entry_fint=0.1, edge_exit_fint=0.2)
mm.edge_entry_active = 1
mm.edge_exit_active = 0
mm.model = 'rot-kick-rot'
mm.num_multipole_kicks = 10


p_test = p0.copy()
p_ref = p0.copy()

mm.track(p_test)
bb.track(p_ref)

xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

# Test backtracking
line = xt.Line(elements=[mm])
line.track(p_test, backtrack=True)
xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)

# ===============================================================================

# Test exit linear edge alone
bb = xt.Bend(h=0.1, k0=0.11, length=0,
             edge_entry_angle=0.02, edge_exit_angle=0.03,
             edge_entry_hgap=0.04, edge_exit_hgap=0.05,
             edge_entry_fint=0.1, edge_exit_fint=0.2)

bb.edge_entry_active = 0
bb.edge_exit_active = 1
bb.model = 'rot-kick-rot'
bb.num_multipole_kicks = 10

mm = Magnet(h=0.1, k0=0.11, length=0,
                edge_entry_angle=0.02, edge_exit_angle=0.03,
                edge_entry_hgap=0.04, edge_exit_hgap=0.05,
                edge_entry_fint=0.1, edge_exit_fint=0.2)

mm.edge_entry_active = 0
mm.edge_exit_active = 1
mm.model = 'rot-kick-rot'
mm.num_multipole_kicks = 10

p_test = p0.copy()
p_ref = p0.copy()

mm.track(p_test)
bb.track(p_ref)

xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

# Test backtracking
line = xt.Line(elements=[mm])
line.track(p_test, backtrack=True)
xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)


# ==============================================================================
# Testa a full bend with linear edges

bb = xt.Bend(h=0.1, k0=0.11, length=10,
                edge_entry_angle=0.02, edge_exit_angle=0.03,
                edge_entry_hgap=0.04, edge_exit_hgap=0.05,
                edge_entry_fint=0.1, edge_exit_fint=0.2)

bb.edge_entry_active = 1
bb.edge_exit_active = 1
bb.model = 'rot-kick-rot'
bb.num_multipole_kicks = 10
bb.edge_entry_model = 'linear'
bb.edge_exit_model = 'linear'

mm = Magnet(h=0.1, k0=0.11, length=10,
                edge_entry_angle=0.02, edge_exit_angle=0.03,
                edge_entry_hgap=0.04, edge_exit_hgap=0.05,
                edge_entry_fint=0.1, edge_exit_fint=0.2)

mm.edge_entry_active = 1
mm.edge_exit_active = 1
mm.model = 'rot-kick-rot'
mm.num_multipole_kicks = 10
mm.edge_entry_model = 'linear'
mm.edge_exit_model = 'linear'

p_test = p0.copy()
p_ref = p0.copy()

mm.track(p_test)
bb.track(p_ref)

xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-13, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

# Test backtracking
line = xt.Line(elements=[mm])
line.track(p_test, backtrack=True)
xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)

# ===============================================================================

# Test entry non-linear edge alone
bb = xt.Bend(h=0.1, k0=0.11, length=0,
             edge_entry_angle=0.02, edge_exit_angle=0.03,
             edge_entry_hgap=0.04, edge_exit_hgap=0.05,
             edge_entry_fint=0.1, edge_exit_fint=0.2)

bb.edge_entry_active = 1
bb.edge_exit_active = 0
bb.model = 'rot-kick-rot'
bb.num_multipole_kicks = 10
bb.edge_entry_model = 'full'

mm = Magnet(h=0.1, k0=0.11, length=0,
                edge_entry_angle=0.02, edge_exit_angle=0.03,
                edge_entry_hgap=0.04, edge_exit_hgap=0.05,
                edge_entry_fint=0.1, edge_exit_fint=0.2)

mm.edge_entry_active = 1
mm.edge_exit_active = 0
mm.model = 'rot-kick-rot'
mm.num_multipole_kicks = 10
mm.edge_entry_model = 'full'

p_test = p0.copy()
p_ref = p0.copy()

mm.track(p_test)
bb.track(p_ref)

xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

# ===============================================================================
# Test exit non-linear edge alone
bb = xt.Bend(h=0.1, k0=0.11, length=0,
             edge_entry_angle=0.02, edge_exit_angle=0.03,
             edge_entry_hgap=0.04, edge_exit_hgap=0.05,
             edge_entry_fint=0.1, edge_exit_fint=0.2)

bb.edge_entry_active = 0
bb.edge_exit_active = 1
bb.model = 'rot-kick-rot'
bb.num_multipole_kicks = 10
bb.edge_exit_model = 'full'

mm = Magnet(h=0.1, k0=0.11, length=0,
                edge_entry_angle=0.02, edge_exit_angle=0.03,
                edge_entry_hgap=0.04, edge_exit_hgap=0.05,
                edge_entry_fint=0.1, edge_exit_fint=0.2)

mm.edge_entry_active = 0
mm.edge_exit_active = 1
mm.model = 'rot-kick-rot'
mm.num_multipole_kicks = 10
mm.edge_exit_model = 'full'

p_test = p0.copy()
p_ref = p0.copy()

mm.track(p_test)
bb.track(p_ref)

xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

# ==============================================================================
# Testa a full bend with non-linear edges

bb = xt.Bend(h=0.1, k0=0.11, length=10,
                edge_entry_angle=0.02, edge_exit_angle=0.03,
                edge_entry_hgap=0.04, edge_exit_hgap=0.05,
                edge_entry_fint=0.1, edge_exit_fint=0.2)

bb.edge_entry_active = 1
bb.edge_exit_active = 1
bb.model = 'rot-kick-rot'
bb.num_multipole_kicks = 10
bb.edge_entry_model = 'full'
bb.edge_exit_model = 'full'

mm = Magnet(h=0.1, k0=0.11, length=10,
                edge_entry_angle=0.02, edge_exit_angle=0.03,
                edge_entry_hgap=0.04, edge_exit_hgap=0.05,
                edge_entry_fint=0.1, edge_exit_fint=0.2)

mm.edge_entry_active = 1
mm.edge_exit_active = 1
mm.model = 'rot-kick-rot'
mm.num_multipole_kicks = 10
mm.edge_entry_model = 'full'
mm.edge_exit_model = 'full'

p_test = p0.copy()
p_ref = p0.copy()

mm.track(p_test)
bb.track(p_ref)

xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=3e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)


# ==============================================================================
# Test a quadrupole with non-linear fringes

qq = xt.Quadrupole(k1=0.11, length=3)
qq.edge_entry_active = 1
qq.edge_exit_active = 1

mm = Magnet(k0=0, k1=0.11, length=3,
            edge_entry_model='full', edge_exit_model='full',
            edge_entry_fint=0.1, edge_exit_fint=0.2, # should be ignored
            edge_entry_hgap=0.04, edge_exit_hgap=0.05) # should be ignored
mm.edge_entry_active = 1
mm.edge_exit_active = 1
mm.model='mat-kick-mat'

p_test = p0.copy()
p_ref = p0.copy()

mm.track(p_test)
qq.track(p_ref)

xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

# ==============================================================================
# Test a sextupole with non-linear fringes

ss = xt.Sextupole(k2=0.11 + 0.03/3., k2s=0.12 -0.04/3, length=3)
ss.edge_entry_active = 1
ss.edge_exit_active = 1

mm = Magnet(k0=0, k2=0.11, k2s=0.12, length=3,
            knl=[0, 0, 0.03], ksl=[0, 0, -0.04],
            edge_entry_model='full', edge_exit_model='full',
            edge_entry_fint=0.1, edge_exit_fint=0.2, # should be ignored
            edge_entry_hgap=0.04, edge_exit_hgap=0.05) # should be ignored
mm.edge_entry_active = 1
mm.edge_exit_active = 1
mm.model = 'drift-kick-drift-expanded'
mm.num_multipole_kicks = 1
mm.integrator = 'uniform'

p_test = p0.copy()
p_ref = p0.copy()

mm.track(p_test)
ss.track(p_ref)

xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

# ==============================================================================
# Test a octupole with non-linear fringes

oo = xt.Octupole(k3=0.3 + 0.02/3., k3s=0.4 - 0.07/3, length=3)
oo.edge_entry_active = 1
oo.edge_exit_active = 1

mm = Magnet(k0=0, k3=0.3, k3s=0.4, length=3,
            knl=[0, 0, 0, 0.02], ksl=[0, 0, 0, -0.07],
            edge_entry_model='full', edge_exit_model='full',
            edge_entry_fint=0.1, edge_exit_fint=0.2, # should be ignored
            edge_entry_hgap=0.04, edge_exit_hgap=0.05) # should be ignored
mm.edge_entry_active = 1
mm.edge_exit_active = 1
mm.model = 'drift-kick-drift-expanded'
mm.num_multipole_kicks = 1
mm.integrator = 'uniform'

p_test = p0.copy()
p_ref = p0.copy()

mm.track(p_test)
oo.track(p_ref)

xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

line = xt.Line(elements=[mm])
line.track(p_test, backtrack=True)
assert np.all(p_test.state == -32)


import numpy as np
from scipy import integrate
from scipy.constants import c as clight
from scipy.constants import e as qe


def beta(s, beta0, alpha_s0):
    '''Beta function in drift space'''
    return beta0-2*alpha_s0*s+(1+alpha_s0**2)/beta0*s**2

def dispersion(s, d0, dp0):
    '''Dispersion in drift space'''
    return d0+s*dp0

def sigma(beta, epsilon0, betagamma):
    '''Betatronic sigma'''
    return np.sqrt(beta*epsilon0/betagamma)

def luminosity(f, nb,
      N1, N2,
      x_1, x_2,
      y_1, y_2,
      px_1, px_2,
      py_1, py_2,
      energy_tot1, energy_tot2,
      deltap_p0_1, deltap_p0_2,
      epsilon_x1, epsilon_x2,
      epsilon_y1, epsilon_y2,
      sigma_z1, sigma_z2,
      beta_x1, beta_x2,
      beta_y1, beta_y2,
      alpha_x1, alpha_x2,
      alpha_y1, alpha_y2,
      dx_1, dx_2,
      dy_1, dy_2,
      dpx_1, dpx_2,
      dpy_1, dpy_2,
      crab_crossing=None,
      verbose=False, sigma_integration=3, rest_mass_b1=0.93827231, rest_mass_b2=0.93827231):
    '''
    Returns luminosity in Hz/cm^2.
    f: revolution frequency
    nb: number of colliding bunch per beam in the specific Interaction Point (IP).
    N1,2: B1,2 number of particle per bunch
    x,y,1,2: horizontal/vertical position at the IP of B1,2, as defined in MADX [m]
    px,y,1,2: px,py at the IP of B1,2, as defined in MADX
    energy_tot1,2: total energy of the B1,2 [GeV]
    deltap_p0_1,2: rms momentum spread of B1,2 (formulas assume Gaussian off-momentum distribution)
    epsilon_x,y,1,2: horizontal/vertical normalized emittances of B1,2 [m rad]
    sigma_z1,2: rms longitudinal spread in z of B1,2 [m]
    beta_x,y,1,2: horizontal/vertical beta-function at IP of B1,2 [m]
    alpha_x,y,1,2: horizontal/vertical alpha-function at IP of B1,2
    dx,y_1,2: horizontal/vertical dispersion-function at IP of B1,2, as defined in MADX [m]
    dpx,y_1,2: horizontal/vertical differential-dispersion-function IP of B1,2, as defined in MADX
    CC_V_x,y,1,2: B1,2 H/V CC total of the cavities that the beam sees before reaching the IP [V]
    CC_f_x,y,1,2: B1,2 H/V CC frequency of cavities that the beam sees before reaching the IP [Hz]
    CC_phase_1,2: B1,2 H/V CC phase of cavities that the beam sees before reaching the IP.
        Sinusoidal function with respect to the center of the bunch is assumed.
        Therefore 0 rad means no kick for the central longitudinal slice [rad]
    RAB_1,2: B1,2 equivalent H/V linear transfer matrix coefficients between the CC
        that the beam sees before reaching the IP and IP itself [SI units]
    verbose: to have verbose output
    sigma_integration: the number of sigma consider for the integration
        (taken into account only if CC(s) is/are present)
    rest_mass_b1, rest_mass_b2: rest mass in GeV
    In MAD-X px is p_x/p_0 (p_x is the x-component of the momentum and p_0 is the design momentum).
    In our approximation we use the paraxial approximation: p_0~p_z so px is an angle.
    Similar arguments holds for py.
    In MAD-X, dx and dpx are the literature dispersion and is derivative in s divided by the relatistic beta.
    In fact, since pt=beta*deltap, where beta is the relativistic Lorentz factor,
    those functions given by MAD-X must be multiplied by beta a number of times equal to the order of
    the derivative to find the functions given in the literature.
    To note that dpx is normalized by the reference momentum (p_s) and not the design momentum (p_0),
    ps = p0(1+deltap). We assume that dpx is the z-derivative of the px.
    '''

    if deltap_p0_1 != 0 or deltap_p0_2 != 0:
        raise ValueError('effect of dispersion not included yet (untested).')

    gamma1 = energy_tot1 / rest_mass_b1
    br_1 = np.sqrt(1-1/gamma1**2)
    betagamma_1 = br_1*gamma1

    gamma2 = energy_tot2 / rest_mass_b2
    br_2 = np.sqrt(1-1/gamma2**2)
    betagamma_2 = br_2*gamma2

    # module of B1 speed
    v0_1=br_1*clight
    # paraxial hypothesis
    vx_1=v0_1*px_1
    vy_1=v0_1*py_1
    vz_1=v0_1*np.sqrt(1-px_1**2-py_1**2)
    v_1=np.array([vx_1, vy_1, vz_1])

    v0_2=br_2*clight # module of B2 speed
    # Assuming counter rotating B2 ('-' sign)
    vx_2=-v0_2*px_2
    vy_2=-v0_2*py_2
    # assuming px_2**2+py_2**2 < 1
    vz_2=-v0_2*np.sqrt(1-px_2**2-py_2**2)
    v_2=np.array([vx_2, vy_2, vz_2])

    if verbose:
        print(f'B1 velocity vector:{v_1}')
        print(f'B2 velocity vector:{v_2}')

    diff_v = v_1-v_2
    cross_v= np.cross(v_1, v_2)

    # normalized to get 1 for the ideal case 
    # NB we assume px_1 and py_1 constant along the z-slices 
    # NOT TRUE FOR CC! In any case the Moeller efficiency is almost equal to 1 in most cases...
    Moeller_efficiency=np.sqrt(clight**2*np.dot(diff_v,diff_v)-np.dot(cross_v,cross_v))/clight**2/2

    def sx1(s):
        '''The sigma_x of B1, quadratic sum of betatronic and dispersive sigma'''
        return np.sqrt(sigma(beta(s, beta_x1, alpha_x1), epsilon_x1, betagamma_1)**2 \
                       + (dispersion(s, br_1*dx_1, br_1*dpx_1)*deltap_p0_1)**2)

    def sy1(s):
        '''The sigma_y of B1, quadratic sum of betatronic and dispersive sigma'''
        return np.sqrt(sigma(beta(s, beta_y1, alpha_y1), epsilon_y1, betagamma_1)**2 \
                       + (dispersion(s, br_1*dy_1, br_1*dpy_1)*deltap_p0_1)**2)

    def sx2(s):
        '''The sigma_x of B2, quadratic sum of betatronic and dispersive sigma'''
        return np.sqrt(sigma(beta(s, beta_x2, alpha_x2), epsilon_x2, betagamma_2)**2 \
                       + (dispersion(s, br_2*dx_2, br_2*dpx_2)*deltap_p0_2)**2)

    def sy2(s):
        '''The sigma_y of B2, quadratic sum of betatronic and dispersive sigma'''
        return np.sqrt(sigma(beta(s, beta_y2, alpha_y2), epsilon_y2, betagamma_2)**2 \
                       + (dispersion(s, br_2*dy_2,  br_2*dpy_2)*deltap_p0_2)**2)

    sigma_z=np.max([sigma_z1, sigma_z2])

    if crab_crossing is not None and 'CC_V_x_1' in crab_crossing.keys():

        raise NotImplementedError('Crab crossing not tested yet')

        CC_V_x_1 = crab_crossing['CC_V_x_1']
        CC_V_y_1 = crab_crossing['CC_V_y_1']
        CC_V_x_2 = crab_crossing['CC_V_x_2']
        CC_V_y_2 = crab_crossing['CC_V_y_2']

        CC_phase_x_1 = crab_crossing['CC_phase_x_1']
        CC_phase_y_1 = crab_crossing['CC_phase_y_1']
        CC_phase_x_2 = crab_crossing['CC_phase_x_2']
        CC_phase_y_2 = crab_crossing['CC_phase_y_2']

        CC_f_x_1 = crab_crossing['CC_f_x_1']
        CC_f_y_1 = crab_crossing['CC_f_y_1']
        CC_f_x_2 = crab_crossing['CC_f_x_2']
        CC_f_y_2 = crab_crossing['CC_f_y_2']

        R12_1 = crab_crossing['R12_1']
        R22_1 = crab_crossing['R22_1']
        R12_2 = crab_crossing['R12_2']
        R22_2 = crab_crossing['R22_2']
        R34_1 = crab_crossing['R34_1']
        R44_1 = crab_crossing['R44_1']
        R34_2 = crab_crossing['R34_2']
        R44_2 = crab_crossing['R44_2']

        def theta_x_1(delta_z):
            # Eq. 3 of https://espace.cern.ch/acc-tec-sector/Chamonix/Chamx2012/papers/RC_9_04.pdf
            return CC_V_x_1/energy_tot1/1e9*np.sin(CC_phase_x_1 + 2*np.pi*CC_f_x_1/clight * delta_z)

        def theta_y_1(delta_z):
            return CC_V_y_1/energy_tot1/1e9*np.sin(CC_phase_y_1 + 2*np.pi*CC_f_y_1/clight * delta_z)

        def theta_x_2(delta_z):
            return CC_V_x_2/energy_tot2/1e9*np.sin(CC_phase_x_2 + 2*np.pi*CC_f_x_2/clight * delta_z)

        def theta_y_2(delta_z):
            return CC_V_y_2/energy_tot2/1e9*np.sin(CC_phase_y_2 + 2*np.pi*CC_f_y_2/clight * delta_z)

        def mx1(s, t):
            '''The mu_x of B1 as straight line'''
            return x_1 + R12_1*theta_x_1(s-clight * t) + (px_1+R22_1*theta_x_1(s-clight * t))*s

        def my1(s, t):
            '''The mu_y of B1 as straight line'''
            return y_1 + R34_1*theta_y_1(s-clight * t) + (py_1+R44_1*theta_y_1(s-clight * t))*s

        def mx2(s, t):
            '''The mu_x of B2 as straight line'''
            return x_2 + R12_2*theta_x_2(s+clight * t) + (px_2+R22_2*theta_x_2(s+clight * t))*s

        def my2(s, t):
            '''The mu_y of B2 as straight line'''
            return y_2 + R34_2*theta_y_2(s+clight * t) + (py_2+R44_2*theta_y_2(s+clight * t))*s

        def kernel_double_integral(t, s):
            return np.exp(0.5*(-(mx1(s, t) - mx2(s, t))**2/(sx1(s)**2 + sx2(s)**2) \
                               -(my1(s, t) - my2(s, t))**2/(sy1(s)**2 + sy2(s)**2) \
                               -(-br_1*clight * t+s)**2/(sigma_z1**2) \
                               -( br_2*clight * t+s)**2/(sigma_z2**2))) \
        /np.sqrt((sx1(s)**2 + sx2(s)**2)*(sy1(s)**2 + sy2(s)**2))/sigma_z1/sigma_z2

        integral=integrate.dblquad((lambda t, s: kernel_double_integral(t, s)),
                                   -sigma_integration*sigma_z, sigma_integration*sigma_z,-sigma_integration*sigma_z/c, sigma_integration*sigma_z/c)
        L0=f*N1*N2*nb * clight/2/np.pi**(2)*integral[0]

    elif crab_crossing is not None and 'phi_crab_x_1' in crab_crossing:

        raise ValueError('Not implemented yet') # Needs testing

        phi_crab_x_1 = crab_crossing['phi_crab_x_1']
        phi_crab_y_1 = crab_crossing['phi_crab_y_1']
        phi_crab_x_2 = crab_crossing['phi_crab_x_2']
        phi_crab_y_2 = crab_crossing['phi_crab_y_2']

        def mx1(s, t):
            '''The mu_x of B1 as straight line'''
            return x_1 + px_1 * s + phi_crab_x_1 * clight * t

        def my1(s, t):
            '''The mu_y of B1 as straight line'''
            return y_1 + py_1 * s + phi_crab_y_1 * clight * t

        # Signs of the crab terms are guessed...
        def mx2(s, t):
            '''The mu_x of B2 as straight line'''
            return x_2 + px_2 * s + phi_crab_x_2 * clight * t

        def my2(s, t):
            '''The mu_y of B2 as straight line'''
            return y_2 + py_2 * s + phi_crab_y_2 * clight * t

        def kernel_double_integral(t, s):
            return np.exp(0.5*(-(mx1(s, t) - mx2(s, t))**2/(sx1(s)**2 + sx2(s)**2) \
                               -(my1(s, t) - my2(s, t))**2/(sy1(s)**2 + sy2(s)**2) \
                               -(-br_1*clight * t+s)**2/(sigma_z1**2) \
                               -( br_2*clight * t+s)**2/(sigma_z2**2))) \
        /np.sqrt((sx1(s)**2 + sx2(s)**2)*(sy1(s)**2 + sy2(s)**2))/sigma_z1/sigma_z2

        integral=integrate.dblquad((lambda t, s: kernel_double_integral(t, s)),
                                   -sigma_integration*sigma_z,
                                   sigma_integration*sigma_z,
                                   -sigma_integration*sigma_z/clight,
                                   sigma_integration*sigma_z/clight)
        L0=f*N1*N2*nb * clight/2/np.pi**(2)*integral[0]

    else:
        def mx1(s):
            '''The mu_x of B1 as straight line'''
            return x_1 + px_1*s

        def my1(s):
            '''The mu_y of B1 as straight line'''
            return y_1 + py_1*s

        def mx2(s):
            '''The mu_x of B2 as straight line'''
            return x_2 + px_2*s

        def my2(s):
            '''The mu_y of B2 as straight line'''
            return y_2 + py_2*s

        def kernel_single_integral(s):
            return np.exp(0.5*(-(mx1(s) - mx2(s))**2/(sx1(s)**2 + sx2(s)**2) \
                               -(my1(s) - my2(s))**2/(sy1(s)**2 + sy2(s)**2) \
                               -((br_1+br_2)**2*s**2)/(br_2**2*sigma_z1**2 + br_1**2*sigma_z2**2))) \
            /np.sqrt((sx1(s)**2 + sx2(s)**2)*(sy1(s)**2 + sy2(s)**2)*(sigma_z1**2 + sigma_z2**2))

        integral=integrate.quad(lambda s: kernel_single_integral(s), -20*sigma_z, 20*sigma_z)
        L0=f*N1*N2*nb/np.sqrt(2)/np.pi**(3/2)*integral[0]
    result= L0*Moeller_efficiency/1e4
    if verbose:
        print(f'Moeller efficiency: {Moeller_efficiency}')
        print(f'Integral Relative Error: {integral[1]/integral[0]}')
        print(f'==> Luminosity [Hz/cm^2]: {result}')
    return result


def luminosity_from_twiss(
    n_colliding_bunches,
    num_particles_per_bunch,
    ip_name,
    nemitt_x,
    nemitt_y,
    sigma_z,
    twiss_b1,
    twiss_b2,
    f_rev=None,
    crab=None):

    assert crab is not None, 'crab crossing information is required'

    twiss_b2_rev = twiss_b2.reverse()

    if crab:
        crab_crossing = {
                'phi_crab_x_1': twiss_b1['dx_zeta', ip_name],
                'phi_crab_x_2': twiss_b2_rev['dx_zeta', ip_name],
                'phi_crab_y_1': twiss_b1['dy_zeta', ip_name],
                'phi_crab_y_2': twiss_b2_rev['dy_zeta', ip_name],
            }
    else:
        crab_crossing = None

    if f_rev is None:
        if 'T_rev0' not in twiss_b1.keys():
            raise ValueError('Revolution frequency cannot be retrieved from twiss, '
                             'please provide f_rev')
        f_rev = 1/twiss_b1.T_rev0

    lumi = luminosity(
        f=f_rev,
        rest_mass_b1=twiss_b1.particle_on_co.mass0 * 1e-9, # GeV
        rest_mass_b2=twiss_b2_rev.particle_on_co.mass0 * 1e-9, # GeV
        nb=n_colliding_bunches,
        N1=num_particles_per_bunch,
        N2=num_particles_per_bunch,
        x_1=twiss_b1['x', ip_name],
        x_2=twiss_b2_rev['x', ip_name],
        y_1=twiss_b1['y', ip_name],
        y_2=twiss_b2_rev['y', ip_name],
        px_1=twiss_b1['px', ip_name],
        px_2=twiss_b2_rev['px', ip_name],
        py_1=twiss_b1['py', ip_name],
        py_2=twiss_b2_rev['py', ip_name],
        energy_tot1=twiss_b1.particle_on_co.energy0[0]*1e-9, # GeV
        energy_tot2=twiss_b2_rev.particle_on_co.energy0[0]*1e-9, # GeV
        deltap_p0_1=0, # energy spread (for now we neglect effect of dispersion)
        deltap_p0_2=0, # energy spread (for now we neglect effect of dispersion)
        epsilon_x1=nemitt_x,
        epsilon_x2=nemitt_x,
        epsilon_y1=nemitt_y,
        epsilon_y2=nemitt_y,
        sigma_z1=sigma_z,
        sigma_z2=sigma_z,
        beta_x1=twiss_b1['betx', ip_name],
        beta_x2=twiss_b2_rev['betx', ip_name],
        beta_y1=twiss_b1['bety', ip_name],
        beta_y2=twiss_b2_rev['bety', ip_name],
        alpha_x1=twiss_b1['alfx', ip_name],
        alpha_x2=twiss_b2_rev['alfx', ip_name],
        alpha_y1=twiss_b1['alfy', ip_name],
        alpha_y2=twiss_b2_rev['alfy', ip_name],
        dx_1=twiss_b1['dx', ip_name],
        dx_2=twiss_b2_rev['dx', ip_name],
        dy_1=twiss_b1['dy', ip_name],
        dy_2=twiss_b2_rev['dy', ip_name],
        dpx_1=twiss_b1['dpx', ip_name],
        dpx_2=twiss_b2_rev['dpx', ip_name],
        dpy_1=twiss_b1['dpy', ip_name],
        dpy_2=twiss_b2_rev['dpy', ip_name],
        crab_crossing=crab_crossing,
        verbose=False, sigma_integration=3)

    return lumi


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
from scipy.constants import c as clight
from scipy.constants import epsilon_0
from scipy.constants import mu_0
from scipy.constants import m_e as me_kg
from scipy.constants import e as qe

me = me_kg*clight**2/qe

from .base_classes import Element
from .be_beamfields.beambeam import BeamBeam4D
from .be_beamfields.beambeam import BeamBeam6D
from .be_beamfields.spacecharge import SCCoasting
from .be_beamfields.spacecharge import SCQGaussProfile
from .be_beamfields.spacecharge import SCInterpolatedProfile

_factorial = np.array(
    [
        1,
        1,
        2,
        6,
        24,
        120,
        720,
        5040,
        40320,
        362880,
        3628800,
        39916800,
        479001600,
        6227020800,
        87178291200,
        1307674368000,
        20922789888000,
        355687428096000,
        6402373705728000,
        121645100408832000,
        2432902008176640000,
    ]
)

class Marker(Element):

    _description = []

    def track(self, p):
        pass

class Drift(Element):
    """Drift in expanded form"""

    _description = [("length", "m", "Length of the drift", 0)]

    def track(self, p):
        length = self.length
        rpp = p.rpp
        xp = p.px * rpp
        yp = p.py * rpp
        p.x += xp * length
        p.y += yp * length
        p.zeta += length * (1. - 1. / p.rvv * (1 + (xp ** 2 + yp ** 2) / 2))
        p.s += length


class DriftExact(Drift):
    """Drift in exact form"""

    _description = [("length", "m", "Length of the drift", 0)]

    def track(self, p):
        sqrt = p._m.sqrt
        length = self.length
        opd = 1 + p.delta
        lpzi = length / sqrt(opd ** 2 - p.px ** 2 - p.py ** 2)
        p.x += p.px * lpzi
        p.y += p.py * lpzi
        p.zeta += length - 1 / p.rvv * opd * lpzi
        p.s += length


def _arrayofsize(ar, size):
    ar = np.array(ar)
    if len(ar) == 0:
        return np.zeros(size, dtype=ar.dtype)
    elif len(ar) < size:
        ar = np.hstack([ar, np.zeros(size - len(ar), dtype=ar.dtype)])
    return ar


class Multipole(Element):
    """ Multipole """

    _description = [
        (
            "knl",
            "m^-n",
            "Normalized integrated strength of normal components",
            lambda: [0],
        ),
        (
            "ksl",
            "m^-n",
            "Normalized integrated strength of skew components",
            lambda: [0],
        ),
        (
            "hxl",
            "rad",
            "Rotation angle of the reference trajectory"
            "in the horizzontal plane",
            0,
        ),
        (
            "hyl",
            "rad",
            "Rotation angle of the reference trajectory in the vertical plane",
            0,
        ),
        ("length", "m", "Length of the originating thick multipole", 0),
    ]

    @property
    def order(self):
        return max(len(self.knl), len(self.ksl)) - 1

    def track(self, p):
        order = self.order
        length = self.length
        knl = _arrayofsize(self.knl, order + 1)
        ksl = _arrayofsize(self.ksl, order + 1)
        x = p.x
        y = p.y
        chi = p.chi
        dpx = knl[order]
        dpy = ksl[order]
        for ii in range(order, 0, -1):
            zre = (dpx * x - dpy * y) / ii
            zim = (dpx * y + dpy * x) / ii
            dpx = knl[ii - 1] + zre
            dpy = ksl[ii - 1] + zim
        dpx = -chi * dpx
        dpy = chi * dpy
        # curvature effect kick
        hxl = self.hxl
        hyl = self.hyl
        delta = p.delta
        if hxl != 0 or hyl != 0:
            b1l = chi * knl[0]
            a1l = chi * ksl[0]
            hxlx = hxl * x
            hyly = hyl * y
            if length > 0:
                hxx = hxlx / length
                hyy = hyly / length
            else:  # non physical weak focusing disabled (SixTrack mode)
                hxx = 0
                hyy = 0
            dpx += hxl + hxl * delta - b1l * hxx
            dpy -= hyl + hyl * delta - a1l * hyy
            p.zeta -= 1. / p.rvv * chi * (hxlx - hyly)
        p.px += dpx
        p.py += dpy


class RFMultipole(Element):
    """
    H= -l sum   Re[ (kn[n](zeta) + i ks[n](zeta) ) (x+iy)**(n+1)/ n ]

    kn[n](z) = k_n cos(2pi w tau + pn/180*pi)
    ks[n](z) = k_n cos(2pi w tau + pn/180*pi)

    """

    _description = [
        ("voltage", "volt", "Voltage", 0),
        ("frequency", "hertz", "Frequency", 0),
        ("lag", "degree", "Delay in the cavity sin(lag - w tau)", 0),
        ("knl", "", "...", lambda: [0]),
        ("ksl", "", "...", lambda: [0]),
        ("pn", "", "...", lambda: [0]),
        ("ps", "", "...", lambda: [0]),
    ]

    @property
    def order(self):
        return max(len(self.knl), len(self.ksl)) - 1

    def track(self, p):
        sin = p._m.sin
        cos = p._m.cos
        pi = p._m.pi
        order = self.order
        k = 2 * pi * self.frequency / clight
        tau = p.zeta / p.beta0
        ktau = k * tau
        deg2rad = pi / 180
        knl = _arrayofsize(self.knl, order + 1)
        ksl = _arrayofsize(self.ksl, order + 1)
        pn = _arrayofsize(self.pn, order + 1) * deg2rad
        ps = _arrayofsize(self.ps, order + 1) * deg2rad
        x = p.x
        y = p.y
        dpx = 0
        dpy = 0
        dptr = 0
        zre = 1
        zim = 0
        for ii in range(order + 1):
            pn_ii = pn[ii] - ktau
            ps_ii = ps[ii] - ktau
            cn = cos(pn_ii)
            sn = sin(pn_ii)
            cs = cos(ps_ii)
            ss = sin(ps_ii)
            # transverse kick order i!
            dpx += cn * knl[ii] * zre - cs * ksl[ii] * zim
            dpy += cs * ksl[ii] * zre + cn * knl[ii] * zim
            # compute z**(i+1)/(i+1)!
            zret = (zre * x - zim * y) / (ii + 1)
            zim = (zim * x + zre * y) / (ii + 1)
            zre = zret
            fnr = knl[ii] * zre
            # fni = knl[ii] * zim
            # fsr = ksl[ii] * zre
            fsi = ksl[ii] * zim
            # energy kick order i+1
            dptr += sn * fnr - ss * fsi

        chi = p.chi
        p.px += -chi * dpx
        p.py += chi * dpy
        dv0 = self.voltage * sin(self.lag * deg2rad - ktau)
        p.add_to_energy(p.charge_ratio * p.q0 * (dv0 - p.p0c * k * dptr))


class Cavity(Element):
    """Radio-frequency cavity"""

    _description = [
        ("voltage", "V", "Integrated energy change", 0),
        ("frequency", "Hz", "Frequency of the cavity", 0),
        ("lag", "degree", "Delay in the cavity sin(lag - w tau)", 0),
    ]

    def track(self, p):
        sin = p._m.sin
        pi = p._m.pi
        k = 2 * pi * self.frequency / clight
        tau = p.zeta / p.beta0
        phase = self.lag * pi / 180 - k * tau
        p.add_to_energy(p.charge_ratio * p.q0 * self.voltage * sin(phase))


class SawtoothCavity(Element):
    """Radio-frequency cavity"""

    _description = [
        ("voltage", "V", "Integrated energy change", 0),
        ("frequency", "Hz", "Equivalent Frequency of the cavity", 0),
        ("lag", "degree", "Delay in the cavity `lag - w tau`", 0),
    ]

    def track(self, p):
        pi = p._m.pi
        k = 2 * pi * self.frequency / clight
        tau = p.zeta / p.beta0
        phase = self.lag * pi / 180 - k * tau
        phase = (phase + pi) % (2 * pi) - pi
        p.add_to_energy(p.charge_ratio * p.q0 * self.voltage * phase)


class XYShift(Element):
    """shift of the reference"""

    _description = [
        ("dx", "m", "Horizontal shift", 0),
        ("dy", "m", "Vertical shift", 0),
    ]

    def track(self, p):
        p.x -= self.dx
        p.y -= self.dy




class Elens(Element):
    """Hollow Electron Lens"""

    _description = [("voltage", "V", "Voltage of the electron lens", 0),
                    ("current", "A", "Current of the e-beam", 0),
                    ("inner_radius", "m", "Inner radius of the hollow e-beam", 0),
                    ("outer_radius", "m", "Outer radius of the hollow e-beam", 0),
                    ("ebeam_center_x", "m", "Center of the e-beam in x", 0),
                    ("ebeam_center_y", "m", "Center of the e-beam in y", 0),
                    ("elens_length", "m", "Length of the hollow electron lens", 0)
                    ]

    def track(self, p):

        # vacuum permittivity
        epsilon0 = epsilon_0
        pi       = np.pi             # pi
        e_mass   = me                # electron mass

        # get the transverse amplitude
        # TO DO: needs to be modified for off-centererd e-beam
        r = np.sqrt(p.x**2 + p.y**2)

        # magnetic rigidity

        if type(p.p0c) is float:
            Brho = p.p0c/(p.q0*clight)
        else:
            Brho = p.p0c[0]/(p.q0*clight)


        # Electron properties
        Ekin_e = self.voltage                         # kinetic energy
        Etot_e = Ekin_e + e_mass                      # total energy
        p_e    = np.sqrt(Etot_e**2 - e_mass**2)       # electron momentum
        beta_e = p_e/Etot_e                           # relativ. beta

        # relativistic beta  of protons
        beta_p = p.rvv*p.beta0

        # abbreviate for better readability
        r1 = self.inner_radius
        r2 = self.outer_radius
        I  = self.current

        # geometric factor frr
        frr = ((r**2 - r1**2)/(r2**2 - r1**2))  # uniform distribution

        try:
            frr = [max(0,iitem) for iitem in frr]
            frr = [min(1,iitem) for iitem in frr]
            frr = np.array(frr, dtype = float)

        except TypeError:
            frr = max(0,frr)
            frr = min(1,frr)
            frr = np.array([frr], dtype=float)


        #
        #
        # if len(frr)>0:
        #     frr[frr<0] = 0
        #     frr[frr>1] = 1

        # calculate the kick at r2 (maximum kick)
        theta_max = ((1/(4*pi*epsilon0))*(2*self.elens_length*I)*
                    (1+beta_e*beta_p)*(1/(r2*Brho*beta_e*beta_p*clight**2)))

        # calculate the kick of the particles
        # the (-1) stems from the attractive force of the E-field
        theta = (-1)*theta_max*r2*p.rpp*p.chi

        print("type frr", type(frr))
        print("type r", type(r))

        theta = theta*np.divide(frr, r, out=np.zeros_like(frr), where=r!=0)

        # convert px and py to x' and y'
        xp   = p.px * p.rpp
        yp   = p.py * p.rpp

        # update xp and yp with the HEL kick
        # use np.divide to not crash when r=0
        xp = xp + p.x*np.divide(theta, r, out=np.zeros_like(theta), where=r!=0)
        yp = yp + p.y*np.divide(theta, r, out=np.zeros_like(theta), where=r!=0)

        # update px and py.
        p.px = xp/p.rpp
        p.py = yp/p.rpp


class Wire(Element):
    """Current-carrying wire"""

    _description = [("L_phy"  ,"m"," Physical length of the wire ",0),
                    ("L_int"  ,"m"," Integration length (embedding drift)",0),
                    ("current","A"," Current in the wire",0),
                    ("xma"    ,"m"," x position of the wire from reference trajectory",0),
                    ("yma"    ,"m"," y position of the wire from reference trajectory",0)
                    ]

    def track(self, p):
        # Data from particle
        x      = p.x
        y      = p.y
        D_x    = x-self.xma
        D_y    = y-self.yma
        R2     = D_x*D_x + D_y*D_y


        # chi = q/q0 * m0/m
        # p0c : reference particle momentum
        # q0  : reference particle charge
        chi    = p.chi
        p0c    = p.p0c
        q0     = p.q0


        # Computing the kick
        L1   = self.L_int + self.L_phy
        L2   = self.L_int - self.L_phy
        N    = mu_0*self.current*q0/(4*np.pi*p0c/clight)

        dpx  =  -N*D_x*(np.sqrt(L1*L1 + 4.0*R2) - np.sqrt(L2*L2 + 4.0*R2))/R2
        dpy  =  -N*D_y*(np.sqrt(L1*L1 + 4.0*R2) - np.sqrt(L2*L2 + 4.0*R2))/R2


        # Update the particle properties
        p.px += dpx
        p.py += dpy



class SRotation(Element):
    """anti-clockwise rotation of the reference frame"""

    _description = [("angle", "", "Rotation angle", 0)]

    def track(self, p):
        deg2rag = p._m.pi / 180
        cz = p._m.cos(self.angle * deg2rag)
        sz = p._m.sin(self.angle * deg2rag)
        xn = cz * p.x + sz * p.y
        yn = -sz * p.x + cz * p.y
        p.x = xn
        p.y = yn
        xn = cz * p.px + sz * p.py
        yn = -sz * p.px + cz * p.py
        p.px = xn
        p.py = yn


class LimitRect(Element):
    _description = [
        ("min_x", "m", "Minimum horizontal aperture", -1.0),
        ("max_x", "m", "Maximum horizontal aperture", 1.0),
        ("min_y", "m", "Minimum vertical aperture", -1.0),
        ("max_y", "m", "Minimum vertical aperture", 1.0),
    ]

    def track(self, particle):

        x = particle.x
        y = particle.y

        if not hasattr(particle.state, "__iter__"):
            particle.state = int(
                x >= self.min_x
                and x <= self.max_x
                and y >= self.min_y
                and y <= self.max_y
            )
        else:
            particle.state = np.int_(
                (x >= self.min_x)
                & (x <= self.max_x)
                & (y >= self.min_y)
                & (y <= self.max_y)
            )
            particle.remove_lost_particles()


class LimitEllipse(Element):
    _description = [
        ("a", "m", "Horizontal semiaxis", 1.0),
        ("b", "m", "Vertical semiaxis", 1.0),
    ]

    def track(self, particle):

        x = particle.x
        y = particle.y

        if not hasattr(particle.state, "__iter__"):
            particle.state = int(
                x * x / (self.a * self.a) + y * y / (self.b * self.b) <= 1.0
            )
        else:
            particle.state = np.int_(
                x * x / (self.a * self.a) + y * y / (self.b * self.b) <= 1.0
            )
            particle.remove_lost_particles()


class LimitRectEllipse(Element):
    _description = [
        ("max_x", "m", "Maximum horizontal aperture", 1.0),
        ("max_y", "m", "Maximum vertical aperture", 1.0),
        ("a", "m", "Horizontal semiaxis", 1.0),
        ("b", "m", "Vertical semiaxis", 1.0),
    ]

    def track(self, particle):

        x = particle.x
        y = particle.y

        if not hasattr(particle.state, "__iter__"):
            particle.state = int(
                x >= -self.max_x
                and x <= self.max_x
                and y >= -self.max_y
                and y <= self.max_y
                and x * x / (self.a * self.a) + y * y / (self.b * self.b) <= 1.0
            )
        else:
            particle.state = np.int_(
                (x >= -self.max_x)
                & (x <= self.max_x)
                & (y >= -self.max_y)
                & (y <= self.max_y)
                & (x * x / (self.a * self.a) + y * y / (self.b * self.b) <= 1.0)
            )
            particle.remove_lost_particles()

class LimitPolygon(Element):
    _description = [
        ("x_vertices", "m", "Horizontal vertices coordinates", ()),
        ("y_vertices", "m", "Vertical vertices coordinates", ()),
    ]

    def track(self, particle):
        raise NotImplementedError

class BeamMonitor(Element):
    _description = [
        ("num_stores", "", "...", 0),
        ("start", "", "...", 0),
        ("skip", "", "...", 1),
        ("max_particle_id", "", "", 0),
        ("min_particle_id", "", "", 0),
        ("is_rolling", "", "", False),
        ("is_turn_ordered", "", "", True),
        ("data", "", "...", lambda: []),
    ]

    def offset(self, particle):
        _offset = -1
        nn = (
            self.max_particle_id >= self.min_particle_id
            and (self.max_particle_id - self.min_particle_id + 1)
            or -1
        )
        assert self.is_turn_ordered

        if (
            particle.turn >= self.start
            and nn > 0
            and particle.particle_id >= self.min_particle_id
            and particle.particle_id <= self.max_particle_id
        ):
            turns_since_start = particle.turns - self.start
            store_index = turns_since_start // self.skip
            if store_index < self.num_stores:
                pass
            elif self.is_rolling:
                store_index = store_index % self.num_stores
            else:
                store_index = -1

            if store_index >= 0:
                _offset = store_index * nn + particle.particle_id

        return _offset

    def track(self, p):
        self.data.append(p.copy)


class DipoleEdge(Element):

    @classmethod
    def from_dict(cls, dct):
        dct = dct.copy()
        for kk in list(dct.keys()):
            if kk == "h" or kk == "_h":
                dct["k"] = dct[kk]
                continue
            if kk.startswith("_"):
                dct[kk[1:]] = dct[kk]
        return super(DipoleEdge, cls).from_dict(dct)

    _description = [
        ("k", "1/m", "Curvature", 0),
        ("e1", "rad", "Face angle", 0),
        ("hgap", "m", "Equivalent gap", 0),
        ("fint", "", "Fringe integral", 0),
    ]

    def track(self, p):
        tan = p._m.tan
        sin = p._m.sin
        cos = p._m.cos
        corr = 2 * self.k * self.hgap * self.fint
        r21 = self.k * tan(self.e1)
        r43 = -self.k * tan(
            self.e1 - corr / cos(self.e1) * (1 + sin(self.e1) ** 2)
        )
        p.px += r21 * p.x
        p.py += r43 * p.y

class LinearTransferMatrix(Element):
    _description = [
        ("alpha_x_0","","",0.0),
        ("beta_x_0","","",0.0),
        ("disp_x_0","","",0.0),
        ("disp_px_0","","",0.0),
        ("alpha_x_1","","",0.0),
        ("beta_x_1","","",0.0),
        ("disp_x_1","","",0.0),
        ("disp_px_1","","",0.0),
        ("alpha_y_0","","",0.0),
        ("beta_y_0","","",0.0),
        ("disp_y_0","","",0.0),
        ("disp_py_0","","",0.0),
        ("alpha_y_1","","",0.0),
        ("beta_y_1","","",0.0),
        ("disp_y_1","","",0.0),
        ("disp_py_1","","",0.0),
        ("Q_x","","",0.0),
        ("Q_y","","",0.0),
        ("beta_s","","",0.0),
        ("Q_s","","",0.0),
        ("bucket_length","","",0.0),
        ("chroma_x","","",0.0),
        ("chroma_y","","",0.0),
        ("det_xx","","",0.0),
        ("det_xy","","",0.0),
        ("det_yy","","",0.0),
        ("det_yx","","",0.0),
        ("energy_ref_increment","","",0.0),
        ("energy_increment","","",0.0),
        ("x_ref_0","","",0.0),
        ("px_ref_0","","",0.0),
        ("x_ref_1","","",0.0),
        ("px_ref_1","","",0.0),
        ("y_ref_0","","",0.0),
        ("py_ref_0","","",0.0),
        ("y_ref_1","","",0.0),
        ("py_ref_1","","",0.0),
        ("damping_rate_x","","",0.0),
        ("damping_rate_px","","",0.0),
        ("damping_rate_y","","",0.0),
        ("damping_rate_py","","",0.0),
        ("damping_rate_zeta","","",0.0),
        ("damping_rate_pzeta","","",0.0),
        ("gauss_noise_ampl_x","","",0.0),
        ("gauss_noise_ampl_px","","",0.0),
        ("gauss_noise_ampl_y","","",0.0),
        ("gauss_noise_ampl_py","","",0.0),
        ("gauss_noise_ampl_zeta","","",0.0),
        ("gauss_noise_ampl_pzeta","","",0.0),
        ("damping_matrix","","",0.0),]

    def track(self,p):
        sin = p._m.sin
        cos = p._m.cos
        sqrt = p._m.sqrt

        cos_s = cos(2.0*np.pi*self.Q_s)
        sin_s = sin(2.0*np.pi*self.Q_s)
        beta_ratio_x =  sqrt(self.beta_x_1/self.beta_x_0)
        beta_prod_x = sqrt(self.beta_x_1*self.beta_x_0)
        beta_ratio_y =  sqrt(self.beta_y_1/self.beta_y_0)
        beta_prod_y = sqrt(self.beta_y_1*self.beta_y_0)

        #Transverse linear uncoupled matrix

        # removing dispersion and close orbit
        old_x=p.x
        old_px=p.px
        old_y=p.y
        old_py=p.py

        p.x -= self.disp_x_0 * p.delta + self.x_ref_0
        p.px -= self.disp_px_0 * p.delta + self.px_ref_0
        p.y -= self.disp_y_0 * p.delta + self.y_ref_0
        p.py -= self.disp_py_0 * p.delta + self.py_ref_0
        # p.zeta += (self.disp_px_0*old_x - self.disp_x_0*old_px + self.disp_py_0*old_y - self.disp_y_0*old_py)/p.rvv

        J_x = 0.5 * (
                (1.0 + self.alpha_x_0*self.alpha_x_0)/self.beta_x_0 * p.x*p.x
                + 2*self.alpha_x_0 * p.x*p.px
                + self.beta_x_0 * p.px*p.px)
        J_y = 0.5 * (
                (1.0 + self.alpha_y_0*self.alpha_y_0)/self.beta_y_0 * p.y*p.y
                + 2*self.alpha_y_0 * p.y*p.py
                + self.beta_y_0 * p.py*p.py)
        phase = 2*np.pi*(self.Q_x+self.chroma_x*p.delta+self.det_xx*J_x+self.det_xy*J_y)
        cos_x = cos(phase)
        sin_x = sin(phase)
        phase = 2*np.pi*(self.Q_y+self.chroma_y*p.delta+self.det_yy*J_y+self.det_yx*J_x)
        cos_y = cos(phase)
        sin_y = sin(phase)

        M00_x = beta_ratio_x*(cos_x+self.alpha_x_0*sin_x)
        M01_x = beta_prod_x*sin_x
        M10_x = ((self.alpha_x_0-self.alpha_x_1)*cos_x
                      -(1+self.alpha_x_0*self.alpha_x_1)*sin_x
                       )/beta_prod_x
        M11_x = (cos_x-self.alpha_x_1*sin_x)/beta_ratio_x
        M00_y = beta_ratio_y*(cos_y+self.alpha_y_0*sin_y)
        M01_y = beta_prod_y*sin_y
        M10_y = ((self.alpha_y_0-self.alpha_y_1)*cos_y
                      -(1+self.alpha_y_0*self.alpha_y_1)*sin_y
                      )/beta_prod_y
        M11_y = (cos_y-self.alpha_y_1*sin_y)/beta_ratio_y

        p.x,p.px = M00_x*p.x + M01_x*p.px, M10_x*p.x + M11_x*p.px
        p.y,p.py = M00_y*p.y + M01_y*p.py, M10_y*p.y + M11_y*p.py

        shift = 0.0
        if self.bucket_length > 0:
            bucket_length_zeta = self.bucket_length*p.beta0*clight
            shift = bucket_length_zeta*np.floor(p.zeta/bucket_length_zeta+0.5);
            p.zeta -= shift
            
        pzeta_new = sin_s*p.zeta/self.beta_s+cos_s*p.pzeta
        zeta_new = cos_s*p.zeta-self.beta_s*sin_s*p.pzeta

        p.zeta = zeta_new + shift
        p.pzeta = pzeta_new

        if self.energy_increment !=0:
            p.add_to_energy(self.energy_increment)

        #Change energy reference
        #In the transverse plane de change is smoothed, i.e. 
        #  both the position and the momentum are scaled,
        #  rather than only the momentum.
        if self.energy_ref_increment != 0:
            old_px = p.px.copy()
            old_py = p.py.copy()
            new_energy0 = p.mass0*p.gamma0 + self.energy_ref_increment
            new_p0c = sqrt(new_energy0*new_energy0-p.mass0*p.mass0)
            new_beta0 = new_p0c / new_energy0
            new_gamma0 = new_energy0 / p.mass0
            geo_emit_factor = sqrt(p.beta0*p.gamma0/new_beta0/new_gamma0)

            p.p0c = new_p0c

            p.x *= geo_emit_factor
            p.px = old_px * geo_emit_factor
            p.y *= geo_emit_factor
            p.py = old_py * geo_emit_factor

        if not hasattr(self.damping_matrix,'__iter__'):
            if (self.damping_rate_x < 0.0 or self.damping_rate_px < 0.0
                    or self.damping_rate_y < 0.0 or self.damping_rate_py < 0.0
                    or self.damping_rate_zeta < 0.0 or self.damping_rate_pzeta < 0.0):
                raise ValueError("Damping rates cannot be negative")
            if self.damping_rate_x > 0.0:
                p.x *= 1.0-self.damping_rate_x
            if self.damping_rate_px > 0.0:
                p.px *= 1.0-self.damping_rate_px
            if self.damping_rate_y > 0.0:
                p.y *= 1.0-self.damping_rate_y
            if self.damping_rate_py > 0.0:
                p.py *= 1.0-self.damping_rate_py
            if self.damping_rate_zeta > 0.0:
                p.zeta *= 1.0-self.damping_rate_zeta
            if self.damping_rate_pzeta > 0.0:
                p.pzeta *= 1.0-self.damping_rate_pzeta
        else:
            assert np.shape(self.damping_matrix) == (6,6)
            transformation = self.damping_matrix + np.identity(6)
            p.x,p.px,p.y,p.py,p.zeta,p.pzeta = transformation[0,0]*p.x+transformation[0,1]*p.px+transformation[0,2]*p.y+transformation[0,3]*p.py+transformation[0,4]*p.zeta+transformation[0,5]*p.pzeta,\
                                               transformation[1,0]*p.x+transformation[1,1]*p.px+transformation[1,2]*p.y+transformation[1,3]*p.py+transformation[1,4]*p.zeta+transformation[1,5]*p.pzeta,\
                                               transformation[2,0]*p.x+transformation[2,1]*p.px+transformation[2,2]*p.y+transformation[2,3]*p.py+transformation[2,4]*p.zeta+transformation[2,5]*p.pzeta,\
                                               transformation[3,0]*p.x+transformation[3,1]*p.px+transformation[3,2]*p.y+transformation[3,3]*p.py+transformation[3,4]*p.zeta+transformation[3,5]*p.pzeta,\
                                               transformation[4,0]*p.x+transformation[4,1]*p.px+transformation[4,2]*p.y+transformation[4,3]*p.py+transformation[4,4]*p.zeta+transformation[4,5]*p.pzeta,\
                                               transformation[5,0]*p.x+transformation[5,1]*p.px+transformation[5,2]*p.y+transformation[5,3]*p.py+transformation[5,4]*p.zeta+transformation[5,5]*p.pzeta


        if (self.gauss_noise_ampl_x < 0.0 or self.gauss_noise_ampl_px < 0.0 
                or self.gauss_noise_ampl_y < 0.0 or self.gauss_noise_ampl_py < 0.0
                or self.gauss_noise_ampl_zeta < 0.0 or self.gauss_noise_ampl_pzeta < 0.0):
            raise ValueError("Noise amplitude cannot be negative")
        if self.gauss_noise_ampl_x > 0.0:
            p.x += self.gauss_noise_ampl_x*np.random.randn(len(p.x))
        if self.gauss_noise_ampl_x > 0.0:
            p.px += self.gauss_noise_ampl_x*np.random.randn(len(p.px))
        if self.gauss_noise_ampl_y > 0.0:
            p.y += self.gauss_noise_ampl_y*np.random.randn(len(p.y))
        if self.gauss_noise_ampl_py > 0.0:
            p.py += self.gauss_noise_ampl_py*np.random.randn(len(p.py))
        if self.gauss_noise_ampl_zeta > 0.0:
            p.zeta += self.gauss_noise_ampl_zeta*np.random.randn(len(p.zeta))
        if self.gauss_noise_ampl_pzeta > 0.0:
            p.pzeta += self.gauss_noise_ampl_pzeta*np.random.randn(len(p.pzeta))

        # re-adding dispersion and closed orbit
        old_x=p.x
        old_px=p.px
        old_y=p.y
        old_py=p.py

        p.x += self.disp_x_1 * p.delta + self.x_ref_1
        p.px += self.disp_px_1 * p.delta + self.px_ref_1 
        p.y += self.disp_y_1 * p.delta + self.y_ref_1
        p.py += self.disp_py_1 * p.delta + self.py_ref_1
        # p.zeta-= (self.disp_px_1*old_x - self.disp_x_1*old_px + self.disp_py_1*old_y - self.disp_y_1*old_py)/p.rvv


class FirstOrderTaylorMap(Element):
    _description = [
        ("length","","",0.0),
        ("m0","","",0.0),
        ("m1","","",0.0)]

    def track(self,p):
        if self.m0 is None:
            self.m0 = np.zeros(6,dtype=np.float64)
        else:
            if len(np.shape(self.m0)) != 1 or np.shape(self.m0)[0] != 6:
                raise ValueError(f'Wrong shape for m0: {np.shape(m0)}')
        if self.m1 is None:
            self.m1 = np.zeros((6,6),dtype=np.float64)
        else:
            if len(np.shape(self.m1)) != 2 or np.shape(self.m1)[0] != 6 or np.shape(self.m1)[1] != 6:
                raise ValueError(f'Wrong shape for m1: {np.shape(m1)}')

        beta0 = p.beta0
        coords0 = np.array([p.x,p.px,p.y,p.py,p.zeta/beta0,p.ptau])
        p.x = self.m0[0] + self.m1[0,0]*coords0[0] + self.m1[0,1]*coords0[1] + self.m1[0,2]*coords0[2] + self.m1[0,3]*coords0[3] + self.m1[0,4]*coords0[4] + self.m1[0,5]*coords0[5]
        p.px = self.m0[1] + self.m1[1,0]*coords0[0] + self.m1[1,1]*coords0[1] + self.m1[1,2]*coords0[2] + self.m1[1,3]*coords0[3] + self.m1[1,4]*coords0[4] + self.m1[1,5]*coords0[5]
        p.y = self.m0[2] + self.m1[2,0]*coords0[0] + self.m1[2,1]*coords0[1] + self.m1[2,2]*coords0[2] + self.m1[2,3]*coords0[3] + self.m1[2,4]*coords0[4] + self.m1[2,5]*coords0[5]
        p.py = self.m0[3] + self.m1[3,0]*coords0[0] + self.m1[3,1]*coords0[1] + self.m1[3,2]*coords0[2] + self.m1[3,3]*coords0[3] + self.m1[3,4]*coords0[4] + self.m1[3,5]*coords0[5]
        tau = self.m0[4] + self.m1[4,0]*coords0[0] + self.m1[4,1]*coords0[1] + self.m1[4,2]*coords0[2] + self.m1[4,3]*coords0[3] + self.m1[4,4]*coords0[4] + self.m1[4,5]*coords0[5]
        ptau = self.m0[5] + self.m1[5,0]*coords0[0] + self.m1[5,1]*coords0[1] + self.m1[5,2]*coords0[2] + self.m1[5,3]*coords0[3] + self.m1[5,4]*coords0[4] + self.m1[5,5]*coords0[5]
        p.delta = np.sqrt(ptau*ptau + 2.0*ptau/p.beta0+1.0)-1.0
        p.zeta = tau * beta0

        if self.length > 0.0:
            raise NotImplementedError('Radiation is not implemented')

__all__ = [
    "BeamBeam4D",
    "BeamBeam6D",
    "BeamMonitor",
    "Cavity",
    "DipoleEdge",
    "Drift",
    "DriftExact",
    "Element",
    "Elens",
    "LimitEllipse",
    "LimitRect",
    "Multipole",
    "RFMultipole",
    "SCCoasting",
    "SCInterpolatedProfile",
    "SCQGaussProfile",
    "SRotation",
    "XYShift",
    "LinearTransferMatrix",
    "FirstOrderTaylorMap"
]



import numpy as np

import xtrack as xt

import matplotlib.pyplot as plt

line = xt.Line.from_json('psb_00_from_mad.json')
line.build_tracker()
line.twiss_default['method'] = '4d'

tw0 = line.twiss()

# inspect a bit...
tw0.qx, tw0.qy

# A few checks on the imported model
line.vars['k0bi1bsw1l11']._info() # Check that the knob controls k0 and the edges

line.element_refs['bi1.bsw1l1.1'].h._info() # Check no reference system curvature

# Build chicane knob (k0)
line.vars['bsw_k0l'] = 0
line.vars['k0bi1bsw1l11'] = (line.vars['bsw_k0l'] / line['bi1.bsw1l1.1'].length)
line.vars['k0bi1bsw1l12'] = (-line.vars['bsw_k0l'] / line['bi1.bsw1l1.2'].length)
line.vars['k0bi1bsw1l13'] = (-line.vars['bsw_k0l'] / line['bi1.bsw1l1.3'].length)
line.vars['k0bi1bsw1l14'] = (line.vars['bsw_k0l'] / line['bi1.bsw1l1.4'].length)

# Inspect:
line.vars['k0bi1bsw1l11']._info()

# Build knob to model eddy currents (k2)
line.vars['bsw_k2l'] = 0
line.element_refs['bi1.bsw1l1.1'].knl[2] = line.vars['bsw_k2l']
line.element_refs['bi1.bsw1l1.2'].knl[2] = -line.vars['bsw_k2l']
line.element_refs['bi1.bsw1l1.3'].knl[2] = -line.vars['bsw_k2l']
line.element_refs['bi1.bsw1l1.4'].knl[2] = line.vars['bsw_k2l']

# Save to file
line.to_json('psb_01_with_chicane.json')


## Checks:

# Match tunes (with chicane off)
line.match(
    targets=[
        xt.Target('qx',  4.4, tol=1e-6),
        xt.Target('qy',  4.45, tol=1e-6)],
    vary=[
        xt.Vary('kbrqf', step=1e-5),
        xt.Vary('kbrqd', step=1e-5)],
)

# Inspect bump and induced beta beating for different bump amplitudes (and no eddy currents)
plt.close('all')
bsw_k0l_ref = 6.6e-2 # Full bump amplitude

bsw_k0l_values = np.linspace(0, bsw_k0l_ref, 5)
fig1 = plt.figure(1)
sp1 = plt.subplot(3,1,1)
sp2 = plt.subplot(3,1,2, sharex=sp1)
sp3 = plt.subplot(3,1,3, sharex=sp1)

colors = plt.cm.rainbow(np.linspace(0,1,len(bsw_k0l_values)))

for ii, vv in enumerate(bsw_k0l_values[::-1]):
    line.vars['bsw_k0l'] = vv
    tw = line.twiss()

    sp1.plot(tw.s, tw.x, color=colors[ii])
    sp2.plot(tw.s, tw.betx, color=colors[ii])
    sp3.plot(tw.s, tw.bety, color=colors[ii])

sp1.set_ylabel('x [m]')
sp2.set_ylabel('betx [m]')
sp3.set_ylabel('bety [m]')
sp3.set_xlabel('s [m]')

# Inspect eddy currents effect (at start chicane ramp)
line.vars['bsw_k0l'] = bsw_k0l_ref * 0.95

bsw_k2l_ref = -9.7429e-2 # Maximum ramp rate

fig2 = plt.figure(2)
sp1 = plt.subplot(3,1,1)
sp2 = plt.subplot(3,1,2, sharex=sp1)
sp3 = plt.subplot(3,1,3, sharex=sp1)
for ii, vv in enumerate([bsw_k2l_ref, 0]):
    line.vars['bsw_k2l'] = vv
    tw = line.twiss()

    sp1.plot(tw.s, tw.x)
    sp2.plot(tw.s, tw.betx)
    sp3.plot(tw.s, tw.bety)

sp1.set_ylabel('x [m]')
sp2.set_ylabel('betx [m]')
sp3.set_ylabel('bety [m]')
sp3.set_xlabel('s [m]')


line.vars['bsw_k2l'] = bsw_k2l_ref
line.vars['bsw_k0l'] = bsw_k0l_ref
assert np.isclose(line['bi1.bsw1l1.1'].knl[2], bsw_k2l_ref, rtol=0, atol=1e-10)
assert np.isclose(line['bi1.bsw1l1.2'].knl[2], -bsw_k2l_ref, rtol=0, atol=1e-10)
assert np.isclose(line['bi1.bsw1l1.3'].knl[2], -bsw_k2l_ref, rtol=0, atol=1e-10)
assert np.isclose(line['bi1.bsw1l1.4'].knl[2], bsw_k2l_ref, rtol=0, atol=1e-10)
assert np.isclose(line['bi1.bsw1l1.1'].k0, bsw_k0l_ref / line['bi1.bsw1l1.1'].length, rtol=0, atol=1e-10)
assert np.isclose(line['bi1.bsw1l1.2'].k0, -bsw_k0l_ref / line['bi1.bsw1l1.2'].length, rtol=0, atol=1e-10)
assert np.isclose(line['bi1.bsw1l1.3'].k0, -bsw_k0l_ref / line['bi1.bsw1l1.3'].length, rtol=0, atol=1e-10)
assert np.isclose(line['bi1.bsw1l1.4'].k0, bsw_k0l_ref / line['bi1.bsw1l1.4'].length, rtol=0, atol=1e-10)

tw = line.twiss()
assert np.isclose(tw['x', 'bi1.tstr1l1'], -0.045716, rtol=0, atol=1e-5)
assert np.isclose(tw['y', 'bi1.tstr1l1'], 0.0000000, rtol=0, atol=1e-5)
assert np.isclose(tw['betx', 'bi1.tstr1l1'], 5.203667, rtol=0, atol=1e-4)
assert np.isclose(tw['bety', 'bi1.tstr1l1'], 6.902887, rtol=0, atol=1e-4)
assert np.isclose(tw.qy, 4.474414126093382, rtol=0, atol=1e-6) # verify that it does not change from one version to the other
assert np.isclose(tw.qx, 4.396717774779403, rtol=0, atol=1e-6)
assert np.isclose(tw.dqy, -8.625637734560598, rtol=0, atol=1e-3)
assert np.isclose(tw.dqx, -3.5604677592626643, rtol=0, atol=1e-3)

line.vars['bsw_k2l'] = bsw_k2l_ref / 3
assert np.isclose(line['bi1.bsw1l1.1'].knl[2], bsw_k2l_ref / 3, rtol=0, atol=1e-10)
assert np.isclose(line['bi1.bsw1l1.2'].knl[2], -bsw_k2l_ref / 3, rtol=0, atol=1e-10)
assert np.isclose(line['bi1.bsw1l1.3'].knl[2], -bsw_k2l_ref / 3, rtol=0, atol=1e-10)
assert np.isclose(line['bi1.bsw1l1.4'].knl[2], bsw_k2l_ref / 3, rtol=0, atol=1e-10)
assert np.isclose(line['bi1.bsw1l1.1'].k0, bsw_k0l_ref / line['bi1.bsw1l1.1'].length, rtol=0, atol=1e-10)
assert np.isclose(line['bi1.bsw1l1.2'].k0, -bsw_k0l_ref / line['bi1.bsw1l1.2'].length, rtol=0, atol=1e-10)
assert np.isclose(line['bi1.bsw1l1.3'].k0, -bsw_k0l_ref / line['bi1.bsw1l1.3'].length, rtol=0, atol=1e-10)
assert np.isclose(line['bi1.bsw1l1.4'].k0, bsw_k0l_ref / line['bi1.bsw1l1.4'].length, rtol=0, atol=1e-10)

tw = line.twiss()
assert np.isclose(tw['x', 'bi1.tstr1l1'], -0.0458633, rtol=0, atol=1e-5)
assert np.isclose(tw['y', 'bi1.tstr1l1'], 0.0000000, rtol=0, atol=1e-5)
assert np.isclose(tw['betx', 'bi1.tstr1l1'], 5.266456, rtol=0, atol=1e-4)
assert np.isclose(tw['bety', 'bi1.tstr1l1'], 6.320286, rtol=0, atol=1e-4)
assert np.isclose(tw.qy, 4.471766776419623, rtol=0, atol=1e-6)
assert np.isclose(tw.qx, 4.398899960718224, rtol=0, atol=1e-6)
assert np.isclose(tw.dqy, -8.2058757683523, rtol=0, atol=1e-3)
assert np.isclose(tw.dqx, -3.563488925077962, rtol=0, atol=1e-3)

# Switch off bsws
line.vars['bsw_k0l'] = 0
line.vars['bsw_k2l'] = 0
assert np.isclose(line['bi1.bsw1l1.1'].knl[2], 0, rtol=0, atol=1e-10)
assert np.isclose(line['bi1.bsw1l1.2'].knl[2], 0, rtol=0, atol=1e-10)
assert np.isclose(line['bi1.bsw1l1.3'].knl[2], 0, rtol=0, atol=1e-10)
assert np.isclose(line['bi1.bsw1l1.4'].knl[2], 0, rtol=0, atol=1e-10)
assert np.isclose(line['bi1.bsw1l1.1'].k0, 0, rtol=0, atol=1e-10)
assert np.isclose(line['bi1.bsw1l1.2'].k0, 0, rtol=0, atol=1e-10)
assert np.isclose(line['bi1.bsw1l1.3'].k0, 0, rtol=0, atol=1e-10)
assert np.isclose(line['bi1.bsw1l1.4'].k0, 0, rtol=0, atol=1e-10)

tw = line.twiss()
assert np.isclose(tw['x', 'bi1.tstr1l1'], 0, rtol=0, atol=1e-5)
assert np.isclose(tw['y', 'bi1.tstr1l1'], 0, rtol=0, atol=1e-5)
assert np.isclose(tw['betx', 'bi1.tstr1l1'], 5.2996347, rtol=0, atol=1e-4)
assert np.isclose(tw['bety', 'bi1.tstr1l1'], 3.838857, rtol=0, atol=1e-4)
assert np.isclose(tw.qy, 4.45, rtol=0, atol=1e-6)
assert np.isclose(tw.qx, 4.4, rtol=0, atol=1e-6)
assert np.isclose(tw.dqy, -7.149781341846406, rtol=0, atol=1e-4)
assert np.isclose(tw.dqx, -3.5655757511587893, rtol=0, atol=1e-4)

plt.show()


import xtrack as xt
import numpy as np

bend = xt.Bend(k0=0.4, h=0.3, length=1,
            # k1=0.1,
            #    shift_x=1e-3, shift_y=2e-3, rot_s_rad=0.2
            )

line = xt.Line(elements=[bend])

line.configure_bend_model(edge='linear', core='expanded')

line.slice_thick_elements(
    slicing_strategies=[xt.Strategy(xt.Teapot(10000))])
line.build_tracker()
line._line_before_slicing.build_tracker()
assert line['e0..995'].parent_name == 'e0'
assert line['e0..995']._parent is line['e0']
assert line['drift_e0..995'].parent_name == 'e0'
assert line['drift_e0..995']._parent is line['e0']
assert line['e0..entry_map'].parent_name == 'e0'
assert line['e0..entry_map']._parent is line['e0']
assert line['e0..exit_map'].parent_name == 'e0'
assert line['e0..exit_map']._parent is line['e0']

p0 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03)
p_ref = p0.copy()
p_slice = p0.copy()

line.track(p_slice)
line._line_before_slicing.track(p_ref)

assert_allclose = np.testing.assert_allclose
assert_allclose = np.testing.assert_allclose
assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-10)
assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-10)
assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-10)
assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-10)
assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-10)
assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-10)

line.to_json('ttt.json')
line2 = xt.Line.from_json('ttt.json')
assert isinstance(line2['e0..995'], xt.ThinSliceBend)
assert line2['e0..995'].parent_name == 'e0'
assert line2['e0..995']._parent is None
assert line2['drift_e0..995'].parent_name == 'e0'
assert line2['drift_e0..995']._parent is None
assert line2['e0..entry_map'].parent_name == 'e0'
assert line2['e0..entry_map']._parent is None
assert line2['e0..exit_map'].parent_name == 'e0'
assert line2['e0..exit_map']._parent is None

line2.build_tracker()
assert isinstance(line2['e0..995'], xt.ThinSliceBend)
assert line2['e0..995'].parent_name == 'e0'
assert line2['e0..995']._parent is line2['e0']
assert isinstance(line2['drift_e0..995'], xt.DriftSliceBend)
assert line2['drift_e0..995'].parent_name == 'e0'
assert line2['drift_e0..995']._parent is line2['e0']
assert isinstance(line2['e0..entry_map'], xt.ThinSliceBendEntry)
assert line2['e0..entry_map'].parent_name == 'e0'
assert line2['e0..entry_map']._parent is line2['e0']
assert isinstance(line2['e0..exit_map'], xt.ThinSliceBendExit)
assert line2['e0..exit_map'].parent_name == 'e0'
assert line2['e0..exit_map']._parent is line2['e0']

line.track(p_slice, backtrack=True)

assert (p_slice.state == 1).all()
assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)

line.optimize_for_tracking()

if bend.shift_x !=0 or bend.shift_y != 0 or bend.rot_s_rad != 0 and bend.k1 != 0:
    assert isinstance(line['e0..995'], xt.Multipole)
else:
    assert isinstance(line['e0..995'], xt.SimpleThinBend)
assert isinstance(line['drift_e0..995'], xt.Drift)

p_slice = p0.copy()
line.track(p_slice)

assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-10)
assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-10)
assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-10)
assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-10)
assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-10)
assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-10)

line.track(p_slice, backtrack=True)

assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)


import numpy as np
from cpymad.madx import Madx

import xtrack as xt

mad = Madx()
mad.call('../../test_data/hllhc15_noerrors_nobb/sequence.madx')
mad.use('lhcb1')

tw_mad_no_coupling = mad.twiss(ripken=True).dframe()

# introduce coupling
mad.sequence.lhcb1.expanded_elements[7].ksl = [0,0.5e-4]

tw_mad_coupling = mad.twiss(ripken=True).dframe()

line = xt.Line.from_madx_sequence(mad.sequence.lhcb1)
line.particle_ref = xt.Particles(p0c=7000e9, mass0=xt.PROTON_MASS_EV)
line.build_tracker()

tw = line.twiss()

bety1 = tw.bety1
betx2 = tw.betx2

twdf = tw.to_pandas()
twdf.set_index('name', inplace=True)

ips = ['ip1', 'ip2', 'ip3', 'ip4', 'ip5', 'ip6', 'ip7', 'ip8']
betx2_at_ips = twdf.loc[ips, 'betx2'].values
bety1_at_ips = twdf.loc[ips, 'bety1'].values

tw_mad_coupling.set_index('name', inplace=True)
beta12_mad_at_ips = tw_mad_coupling.loc[[ip+':1' for ip in ips], 'beta12'].values
beta21_mad_at_ips = tw_mad_coupling.loc[[ip+':1' for ip in ips], 'beta21'].values

assert np.allclose(betx2_at_ips, beta12_mad_at_ips, rtol=1e-4, atol=0)
assert np.allclose(bety1_at_ips, beta21_mad_at_ips, rtol=1e-4, atol=0)

cmin = tw.c_minus

ttmm = mad.twiss(ripken=True)
dqmin_phase_mad = ttmm.summary.dqmin_phase

tt = line.twiss(strengths=True)
WW = tw.W_matrix[0]
# https://www.agsrhichome.bnl.gov//AP/ap_notes/ap_note_174.pdf

k1sl = tt['k1sl']
c_min = 1 / (2*np.pi) * np.sum(k1sl * np.sqrt(tt.betx * tt.bety)
                        * np.exp(1j * 2 * np.pi * (tt.mux - tt.muy)))
# assert np.isclose(cmin, mad.table.summ.dqmin[0], rtol=0, atol=1e-5)

# Eq. 9 in https://www.rhichome.bnl.gov/AP/ap_notes/ap_note_174.pdf
r1 = np.sqrt(WW[2, 0]**2 + WW[2, 1]**2) / WW[0, 0]
r2 = np.sqrt(WW[0, 2]**2 + WW[0, 3]**2) / WW[2, 2]
Delta = np.mod(tt.qx - tt.qy, 1)

# Eq. 10 in https://www.rhichome.bnl.gov/AP/ap_notes/ap_note_174.pdf
phi = np.angle(WW[2, 0] + 1j*WW[2, 1])

WW_ip7 = tw['W_matrix', 'ip7']
phi_ip7 = np.angle(WW_ip7[2, 0] + 1j*WW_ip7[2, 1])
tt_ip7 = line.twiss(zero_at='ip7', strengths=True)
cmin_ip7 = 1 / (2*np.pi) * np.sum(tt_ip7.k1sl * np.sqrt(tt_ip7.betx * tt_ip7.bety)
                        * np.exp(1j * 2 * np.pi * (tt_ip7.mux - tt_ip7.muy)))

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
sp1 = plt.subplot(211)
plt.plot(tw.s, tw.bety1, label='Xsuite')
plt.plot(tw_mad_coupling.s, tw_mad_coupling.beta21, '--', label='Madx')
plt.ylabel(r'$\beta_{1,y}$')
plt.legend(loc='upper right')
plt.subplot(212, sharex=sp1)
plt.plot(tw.s, tw.betx2, label='betx2')
plt.plot(tw_mad_coupling.s, tw_mad_coupling.beta12, '--')
plt.ylabel(r'$\beta_{2,x}$')
plt.suptitle(r'Xsuite: $C^{-}$'
             f" = {cmin:.2e} "
             r"MAD-X: $C^{-}$ = "
             f"{mad.table.summ.dqmin[0]:.2e}"
             )


plt.show()